from io import BytesIO
from matplotlib.figure import Figure
from matplotlib import pyplot as plt
import numpy as np
import pickle
from liquer.commands import command
from liquer.state_types import StateType, register_state_type
from liquer.constants import mimetype_from_extension

import pickle


class MatplotlibFigureStateType(StateType):
    def identifier(self):
        return "matplotlibfigure"

    def default_extension(self):
        return "pickle"

    def is_type_of(self, data):
        return isinstance(data, plt.Figure)

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)
        if extension in ("pkl", "pickle"):
            output = BytesIO()
            pickle.dump(data, output)
            return output.getvalue(), mimetype
        elif extension in ("png", "svg", "pdf", "ps", "eps", "svgz"):
            output = BytesIO()
            data.savefig(output, dpi=300, format=extension)
            return output.getvalue(), mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by Matplotlib Figure type."
            )

    def from_bytes(self, b: bytes, extension=None):
        if extension is None:
            extension = self.default_extension()
        f = BytesIO()
        f.write(b)
        f.seek(0)

        if extension in ("pickle", "pkl"):
            return pickle.loads(b)
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by Matplotlib Figure type."
        )

    def copy(self, data):
        return data

    def data_characteristics(self, data):
        return dict(description=f"Matplotlib figure")


MATPLOTLIB_FIGURE_STATE_TYPE = MatplotlibFigureStateType()
register_state_type(plt.Figure, MATPLOTLIB_FIGURE_STATE_TYPE)


@command
def mpl(state, *series):
    """Matplotlib chart"""
    fig = plt.figure(figsize=(8, 6), dpi=300)
    axis = fig.add_subplot(1, 1, 1)
    series = list(reversed(list(series)))
    df = state.get()
    extension = None

    while len(series):
        t = series.pop()
        if t in ["jpg", "png", "svg"]:
            extension = t
            continue
        elif t == "xy":
            xcol = series.pop()
            ycol = series.pop()
            state.log_info(f"Chart XY ({xcol} {ycol})")
            axis.plot(df[xcol], df[ycol], label=ycol)
            continue
        elif t == "xye":
            xcol = series.pop()
            ycol = series.pop()
            ecol = series.pop()
            state.log_info(f"Chart XY ({xcol} {ycol}) Error:{ecol}")
            axis.errorbar(df[xcol], df[ycol], yerr=df[ecol], label=ycol)
            continue
        elif t == "xyee":
            xcol = series.pop()
            ycol = series.pop()
            e1col = series.pop()
            e2col = series.pop()
            state.log_info(f"Chart XY ({xcol} {ycol}) Error:({e1col},{e2col})")
            axis.errorbar(df[xcol], df[ycol], yerr=[df[e1col], df[e2col]], label=ycol)
            continue
        elif t == "cxy":
            c = series.pop()
            xcol = series.pop()
            ycol = series.pop()
            axis.plot(df[xcol], df[ycol], c, label=ycol)
            continue
        else:
            state.log_warning(f"Unrecognized MPL parameter {t}")
    # fig.legend()
    if extension is None:
        return fig
    else:
        output = BytesIO()
        fig.savefig(output, dpi=300, format=extension)
        return state.with_data(output.getvalue()).with_filename(f"image.{extension}")



from io import StringIO, BytesIO
from liquer.state_types import StateType, register_state_type
from liquer.constants import mimetype_from_extension
from liquer.commands import command, first_command
from liquer.context import get_context
from liquer.recipes import Recipe, register_recipe
from liquer.parser import parse
from liquer.store import key_extension, key_name, key_name_without_extension
from tempfile import TemporaryDirectory
from pathlib import Path

import pandas as pd
import datafusion as daf
import pyarrow
import pyarrow.parquet
import pyarrow.csv
import pyarrow.feather


class DatafusionDataframeStateType(StateType):
    def identifier(self):
        return "datafusion_dataframe"

    def default_extension(self):
        return "parquet"

    def is_type_of(self, data):
        return isinstance(data, daf.DataFrame)

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)

        if extension == "parquet":
            with TemporaryDirectory() as tmpdir:
                path = Path(tmpdir) / f"data.{extension}"
                table = pyarrow.Table.from_batches(data.collect())
                pyarrow.parquet.write_table(table, str(path))
                b = path.read_bytes()
                return b, mimetype
        elif extension == "csv":
            with TemporaryDirectory() as tmpdir:
                path = Path(tmpdir) / f"data.{extension}"
                table = pyarrow.Table.from_batches(data.collect())
                pyarrow.csv.write_csv(table, str(path))
                b = path.read_bytes()
                return b, mimetype
        elif extension == "feather":
            with TemporaryDirectory() as tmpdir:
                path = Path(tmpdir) / f"data.{extension}"
                table = pyarrow.Table.from_batches(data.collect())
                pyarrow.feather.write_feather(table, str(path))
                b = path.read_bytes()
                return b, mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by DataFusion data-frame type."
            )

    def from_bytes(self, b: bytes, extension=None):
        raise Exception(
            f"Deserialization is not supported by DataFusion data-frame type."
        )

    def copy(self, data):
        return self.from_bytes(self.as_bytes(data)[0])

    def data_characteristics(self, data):
        return dict(description=f"DataFusion data-frame")


DATAFUSION_DATAFRAME_STATE_TYPE = DatafusionDataframeStateType()
register_state_type(daf.DataFrame, DATAFUSION_DATAFRAME_STATE_TYPE)


class DatafusionContextStateType(StateType):
    def identifier(self):
        return "datafusion_context"

    def default_extension(self):
        return "pickle"

    def is_type_of(self, data):
        return isinstance(data, daf.ExecutionContext)

    def as_bytes(self, data, extension=None):
        raise Exception(
            f"Serialization is not supported by DataFusion ExecutionContext type."
        )

    def from_bytes(self, b: bytes, extension=None):
        raise Exception(
            f"Deserialization is not supported by DataFusion ExecutionContext type."
        )

    def copy(self, data):
        raise Exception(
            f"Copy is not supported by DataFusion ExecutionContext type."
        )

    def data_characteristics(self, data):
        return dict(description=f"DataFusion execution context")


DATAFUSION_CONTEXT_STATE_TYPE = DatafusionContextStateType()
register_state_type(daf.ExecutionContext, DATAFUSION_CONTEXT_STATE_TYPE)


@command
def datafusion_to_pandas(df):
    """Convert DataFusion data-frame to pandas data-frame"""
    table = pyarrow.Table.from_batches(df.collect())
    return table.to_pandas()


class ParquetSQLRecipe(Recipe):
    @classmethod
    def recipe_type(self):
        return "parquet_sql"

    @classmethod
    def from_dict(cls, d):
        return cls(d)

    def metadata(self, key):
        metadata = {}
        if "title" in self.data:
            metadata["title"] = self.data["title"]
        if "description" in self.data:
            metadata["description"] = self.data["description"]
        return metadata

    def provides(self):
        if "filename" not in self.data:
            raise Exception(
                f"Recipe {self.recipe_name()} of type {self.recipe_type()} does not have a filename.")
        return self.data.get("provides", [self.data["filename"]])

    def make_execution_context(self, tmpdir, store, context):
        import datafusion as daf
        ctx = daf.ExecutionContext()
        register = self.data.get("register", [])
        store = store.root_store()

        path = Path(tmpdir)
        for query in register:
            try:
                q = parse(query)
            except:
                context.warning(
                    f"Could not parse query '{query}' in parquet_sql recipe {self.recipe_name()}", traceback=traceback.format_exc())
            if q.is_resource_query():
                key = q.resource_query().path()
                if store.is_dir(key):
                    for k in store.listdir_keys(key):
                        if not store.is_dir(k) and key_extension(k) == "parquet":
                            (path / key_name(k)).write_bytes(store.get_bytes(k))
                            context.info(
                                f"Registering {key_name_without_extension(k)} from {key}")
                            ctx.register_parquet(key_name_without_extension(
                                k), str(path / key_name(k)))
                else:
                    (path / key_name(key)).write_bytes(store.get_bytes(key))
                    context.info(f"Registering {key}")
                    ctx.register_parquet(key_name_without_extension(
                        key), str(path / key_name(key)))
            else:
                filename = q.filename()
                if filename is None:
                    context.warning(
                        f"Skipping '{query}' registering because it is lacking a filename")
                    continue
                v = filename.split(".")
                context.info(f"Evaluating query {query}")
                context.evaluate_and_save(
                    query, target_directory=str(tmpdir), target_file=filename)
                context.info(f"Registering {v[0]} from query {query}")
                ctx.register_parquet(v[0], str(path / filename))
        return ctx

    def make(self, key, store=None, context=None):
        context = get_context(context)
        if "sql" not in self.data:
            raise Exception(
                f"Recipe {self.recipe_name()} of type {self.recipe_type()} does not have sql.")
        if "filename" not in self.data:
            raise Exception(
                f"Recipe {self.recipe_name()} of type {self.recipe_type()} does not have a filename.")
        if store is None:
            store = context.store()
        with TemporaryDirectory() as tmpdir:
            ctx = self.make_execution_context(tmpdir, store, context)
            df = ctx.sql(self.data["sql"])
            table = pyarrow.Table.from_batches(df.collect())
            path = Path(tmpdir) / self.data["filename"]
            pyarrow.parquet.write_table(table, str(path))
            b = path.read_bytes()
            metadata = self.metadata(key)
            store.store(key, b, metadata)

register_recipe(ParquetSQLRecipe)


from datetime import datetime


def format_datetime(dt):
    return dt.strftime("%Y-%m-%d %H:%M:%S")

def format_time(dt):
    return dt.strftime("%H:%M:%S")

def now():
    return format_datetime(datetime.now())

def timestamp():
    date = datetime.now()
    return date.isoformat()

def to_datetime(x):
    if isinstance(x, datetime):
        return x
    if isinstance(x, str):
        try:
            return datetime.fromisoformat(x)
        except:
            return datetime.strptime(x, "%Y-%m-%d %H:%M:%S")



import tornado.ioloop
import tornado.web
import liquer.server.handlers as h


def liquer_static_path():
    import liquer.server
    import os.path

    return os.path.join(os.path.dirname(liquer.server.__file__), "static")


class LiquerIndexHandler(h.LiquerIndexHandler, tornado.web.RequestHandler):
    pass

class LiquerJsHandler(h.LiquerJsHandler, tornado.web.RequestHandler):
    pass

# /api/commands.json
class CommandsHandler(h.CommandsHandler, tornado.web.RequestHandler):
    pass

#'/submit/<query>
class SubmitHandler(h.SubmitHandler, tornado.web.RequestHandler):
    pass

# /q/<path:query>
class QueryHandler(h.QueryHandler, tornado.web.RequestHandler):
    pass

#/api/cache/get/<path:query>
class CacheGetDataHandler(h.CacheGetDataHandler, tornado.web.RequestHandler):
    pass

#/api/cache/meta/<path:query>
class CacheMetadataHandler(h.CacheMetadataHandler, tornado.web.RequestHandler):
    pass

#/api/cache/remove/<path:query>
class CacheRemoveHandler(h.CacheRemoveHandler, tornado.web.RequestHandler):
    pass

#/api/cache/contains/<path:query>
class CacheContainsHandler(h.CacheContainsHandler, tornado.web.RequestHandler):
    pass

#/api/cache/keys.json
class CacheKeysHandler(h.CacheKeysHandler, tornado.web.RequestHandler):
    pass

#/api/cache/clean
class CacheCleanHandler(h.CacheCleanHandler, tornado.web.RequestHandler):
    pass

#/api/commands.json
class CommandsHandler(h.CommandsHandler, tornado.web.RequestHandler):
    pass

#/api/debug-json/<path:query>
class GetMetadataHandler(h.GetMetadataHandler, tornado.web.RequestHandler):
    pass

#/api/stored_metadata/<path:query>
class GetStoredMetadataHandler(h.GetStoredMetadataHandler, tornado.web.RequestHandler):
    pass

# /api/build
class BuildHandler(h.BuildHandler, tornado.web.RequestHandler):
    pass

#'/api/register_command/
class RegisterCommandHandler(h.RegisterCommandHandler, tornado.web.RequestHandler):
    pass

#/api/store/data/<path:query>
class StoreDataHandler(h.StoreDataHandler, tornado.web.RequestHandler):
    pass

#/api/store/metadata/<path:query>
class StoreMetadataHandler(h.StoreMetadataHandler, tornado.web.RequestHandler):
    pass

#/web/<path:query>
class WebStoreHandler(h.WebStoreHandler, tornado.web.RequestHandler):
    pass

#/api/store/remove/<path:query>
class StoreRemoveHandler(h.StoreRemoveHandler, tornado.web.RequestHandler):
    pass

#/api/store/removedir/<path:query>
class StoreRemovedirHandler(h.StoreRemovedirHandler, tornado.web.RequestHandler):
    pass

#/api/store/contains/<path:query>
class StoreContainsHandler(h.StoreContainsHandler, tornado.web.RequestHandler):
    pass

#/api/store/is_dir/<path:query>
class StoreIsDirHandler(h.StoreIsDirHandler, tornado.web.RequestHandler):
    pass

#/api/store/keys")
class StoreKeysHandler(h.StoreKeysHandler, tornado.web.RequestHandler):
    pass

#/api/store/listdir/<path:query>
class StoreListdirHandler(h.StoreListdirHandler, tornado.web.RequestHandler):
    pass

#/api/store/makedir/<path:query>
class StoreMakedirHandler(h.StoreMakedirHandler, tornado.web.RequestHandler):
    pass


#class DebugQueryHandler(h.DebugQueryHandler, tornado.web.RequestHandler):
#    pass

def url_mapping():
    return [
        (r"/liquer/api/commands.json", CommandsHandler),
        (r"/liquer/submit/(.*)", SubmitHandler),
        (r"/liquer/q/(.*)", QueryHandler),
        (r"/liquer/api/cache/get/(.*)", CacheGetDataHandler),
        (r"/liquer/api/cache/meta/(.*)", CacheMetadataHandler),
        (r"/liquer/api/cache/remove/(.*)", CacheRemoveHandler),
        (r"/liquer/api/cache/contains/(.*)", CacheContainsHandler),
        (r"/liquer/api/cache/keys.json", CacheKeysHandler),
        (r"/liquer/api/cache/clean", CacheCleanHandler),
        (r"/liquer/api/commands.json", CommandsHandler),
        (r"/liquer/api/debug-json/(.*)", GetMetadataHandler),
        (r"/liquer/api/metadata/(.*)", GetMetadataHandler),
        (r"/liquer/api/stored_metadata/(.*)", GetStoredMetadataHandler),
        (r"/liquer/api/build", BuildHandler),
        (r"/liquer/api/register_command", RegisterCommandHandler),
        (r"/liquer/api/store/data/(.*)", StoreDataHandler),
        (r"/liquer/api/store/metadata/(.*)", StoreMetadataHandler),
        (r"/liquer/web/(.*)", WebStoreHandler),
        (r"/liquer/api/store/remove/(.*)", StoreRemoveHandler),
        (r"/liquer/api/store/removedir/(.*)", StoreRemovedirHandler),
        (r"/liquer/api/store/contains/(.*)", StoreContainsHandler),
        (r"/liquer/api/store/is_dir/(.*)", StoreIsDirHandler),
        (r"/liquer/api/store/keys", StoreKeysHandler),
        (r"/liquer/api/store/listdir/(.*)", StoreListdirHandler),
        (r"/liquer/api/store/makedir/(.*)", StoreMakedirHandler),
    ]

if __name__ == "__main__":
    import liquer.ext.basic
    import liquer.ext.meta
    import liquer.ext.lq_pandas

    #    import liquer.ext.lq_hxl
    import liquer.ext.lq_python
    import liquer.ext.lq_pygments
    from liquer.state import set_var, get_vars

    #    import liquer.commands
    #    liquer.commands.enable_remote_registration()

    url_prefix = "/liquer"
    port = 8080
    set_var("api_path", url_prefix + "/q/")
    set_var("server", f"http://localhost:{port}")

    application = tornado.web.Application(
        url_mapping() + [
            (
                r"/static/(.*)",
                tornado.web.StaticFileHandler,
                {"path": h.liquer_static_path()},
            ),
            (
                r"/liquer/static/(.*)",
                tornado.web.StaticFileHandler,
                {"path": h.liquer_static_path()},
            ),
            (r"/", LiquerIndexHandler),
            (r"/index.html", LiquerIndexHandler),
            (r"/liquer.js", LiquerJsHandler),
        ]
    )
    application.listen(port)
    tornado.ioloop.IOLoop.current().start()



from enum import Enum


class Status(Enum):
    """Status of evaluated data.
    NONE         - status not available (default value)
    SUBMITTED    - immediately after submission (e.g. with pool.evaluate_in_background)
    EVALUATING_PARENT - parent is being evaluated
    EVALUATION   - last action of the query is being evaluated
    EVALUATING_DEPENDENCIES - evaluating dependencies, e.g. parameters
    READY        - data is ready (after successful evaluation)
    ERROR        - evaluation terminated with an error
    EXPIRED      - data exists, but it is obsolete or expired
    RECIPE       - data are not in store, but a recipe is available
    EXTERNAL     - data is from external source or modified by a user
    SIDE_EFFECT  - data has been created as a side-effect of a query
    
    """
    NONE = "none" # If status does not exist for some reason
    SUBMITTED = "submitted" # Immediately after submission
    EVALUATING_PARENT = "parent"
    EVALUATION = "evaluation"
    EVALUATING_DEPENDENCIES = "dependencies"
    ERROR = "error"
    RECIPE = "recipe"
    READY = "ready"
    EXPIRED = "expired"
    EXTERNAL = "external"
    SIDE_EFFECT = "side-effect"

MIMETYPES = dict(
    json="application/json",
    djson="application/json",
    js="text/javascript",
    txt="text/plain",
    html="text/html",
    htm="text/html",
    md="text/markdown",
    xls="application/vnd.ms-excel",
    xlsx="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    pptx="application/vnd.openxmlformats-officedocument.presentationml.presentation",
    ods="application/vnd.oasis.opendocument.spreadsheet",
    tsv="text/tab-separated-values",
    csv="text/csv",
    css="text/css",
    yaml="application/x-yaml",
    msgpack="application/x-msgpack",
    hdf5="application/x-hdf",
    h5="application/x-hdf",
    b="application/octet-stream",
    bin="application/octet-stream",
    pkl="application/octet-stream",
    pickle="application/octet-stream",
    parquet="application/octet-stream",
    feather="application/octet-stream",
    wasm="application/wasm",
    mid="audio/midi",
    midi="audio/midi",
    pdf="application/pdf",
    ps="application/postscript",
    eps="image/eps",
    png="image/png",
    svg="image/svg+xml",
    jpg="image/jpeg",
    jpeg="image/jpeg",
    tga="image/tga",
    tiff="image/tiff",
    bmp="image/bmp",
    pcx="image/x-pcx",
    ppm="image/x-portable-pixmap",
    xbm="image/x-xbitmap",
    ico="image/x-icon",
    gif="image/gif",
    pcd="application/octet-stream",
    psd="application/octet-stream",
    wmf="image/wmf",
    aiff="audio/aiff",
    avi="video/avi",
)

TYPE_IDENTIFIER_BY_EXTENSION = dict(
    json="generic",
    djson="dictionary",
    js="text",
    txt="text",
    html="text",
    htm="text",
    md="text",
    yaml="text",
    xlsx="openpyxl_workbook",
    pptx="pptx_presentation",
    tsv="text",
    csv="dataframe",
    css="text",
    svg="text",
    b="bytes",
    pkl="pickle",
    pickle="pickle",
    parquet="dataframe",
    feather="dataframe",
    eps="image",
    png="image",
    jpg="image",
    jpeg="image",
    tga="image",
    tiff="image",
    bmp="image",
    pcx="image",
    ppm="image",
    xbm="image",
    ico="image",
    gif="image",
    pcd="image",
    psd="image",
    wmf="image",
)

def mimetype_from_extension(extension, default = "application/octet-stream"):
    return MIMETYPES.get(extension, default)

def type_identifier_from_extension(extension):
    return TYPE_IDENTIFIER_BY_EXTENSION.get(extension)



from liquer.constants import Status
from liquer.util import timestamp
from liquer.dependencies import Dependencies
from copy import deepcopy
import json


class Metadata:
    """Metadata wrapper
    Highlevel getters/setters and validation on top of a metadata dictionary object.
    Metadata dictionary is a simple JSON-able python dictionary.
    The Metadata class takes care that all required fields are filled in in the right format.
    High-level logging functionality into the metadata is provided as well. 
    """

    def __init__(self, metadata=None):
        if metadata is None:
            metadata={}
        self.set_metadata(metadata)

    @classmethod
    def from_string(cls, string):
        return cls(json.loads(string))

    def set_metadata(self, metadata):
        """Fills the missing fields and sets the metadata from a dictionary"""
        if "log" not in metadata:
            metadata["log"] = []
        if "child_log" not in metadata:
            metadata["child_log"] = []
        if "query" not in metadata:
            metadata["query"] = None
        if "status" not in metadata:
            metadata["status"] = Status.NONE.value
        if "type_identifier" not in metadata:
            metadata["type_identifier"] = None
        if "message" not in metadata:
            metadata["message"] = ""
        if "is_error" not in metadata:
            metadata["is_error"] = False
        metadata["dependencies"] = Dependencies(metadata.get("dependencies",dict(query=metadata["query"]))).as_dict()
        self.metadata = metadata
        return self

    def get(self, key, default=None):
        return self.metadata.get(key, default)

    @property
    def query(self):
        return self.metadata["query"]

    @query.setter
    def query(self, value):
        self.metadata["query"] = value
        self.metadata["dependencies"]["query"] = value

    @property
    def key(self):
        return self.metadata.get("key")

    @key.setter
    def key(self, value):
        self.metadata["key"] = value

    @property
    def status(self):
        return self.metadata["status"]

    @status.setter
    def status(self, value):
        if isinstance(value, Status):
            value = value.value
        self.metadata["status"] = value
        if value == Status.ERROR.value:
            self.metadata["is_error"] = True

    @property
    def type_identifier(self):
        return self.metadata["type_identifier"]

    @type_identifier.setter
    def type_identifier(self, value):
        self.metadata["type_identifier"] = value

    @property
    def message(self):
        return self.metadata["message"]

    @message.setter
    def message(self, value):
        self.metadata["message"] = value

    @property
    def is_error(self):
        return self.metadata["is_error"]

    @query.setter
    def is_error(self, value):
        value = bool(value)
        self.metadata["is_error"] = value
        if value:
            self.metadata["status"] = Status.ERROR.value

    @property
    def log(self):
        return self.metadata["log"]

    def as_dict(self):
        """Represent metadata as a dictionary - suitable for json serialization
        State data is NOT part of the returned dictionary.
        """
        return deepcopy(self.metadata)

    def add_command_dependency(self, ns, command_metadata, detect_collisions=True):
        dependencies = Dependencies(self.metadata["dependencies"])
        dependencies.add_command_dependency(ns, command_metadata, detect_collisions)
        self.metadata["dependencies"]=dependencies.as_dict()
        return self

    def log_dict(self, d):
        "Put dictionary with a log entry into the log"
        d["timestamp"] = timestamp()
        self.log.append(d)
        if "message" in d:
            self.message = d["message"]
        return self

    def error(self, message, position=None, query=None):
        """Log an error message"""
        self.status = Status.ERROR
        return self.log_dict(
            dict(
                kind="error",
                message=message,
                position=None if position is None else position.to_dict(),
                query=query,
            )
        )

    def warning(self, message, traceback=None):
        """Log a warning message"""
        return self.log_dict(dict(kind="warning", message=message, traceback=traceback))

    def exception(self, message, traceback, position=None, query=None):
        """Log an exception"""
        self.status = Status.ERROR
        return self.log_dict(
            dict(
                kind="error",
                message=message,
                traceback=traceback,
                position=None if position is None else position.to_dict(),
                query=query,
            )
        )

    def info(self, message):
        """Log a message (info)"""
        self.log_dict(dict(kind="info", message=message))
        return self

    def debug(self, message):
        """Log a message (debug)"""
        self.log_dict(dict(kind="debug", message=message))
        return self



from liquer.parser import *
from liquer.store import get_store, KeyNotFoundStoreException, set_store
from liquer.cache import get_cache

def get_stored_metadata(query):
    """Get metadata for a query - if it is stored in cache or a store.    
    """
    if not isinstance(query,Query):
        query = parse(query)
    if query.is_resource_query():
        rq = query.resource_query()
        header = rq.header
        key = rq.path()
        try:
            return get_store().get_metadata(key)
        except KeyNotFoundStoreException:
            return None
    else:
        return get_cache().get_metadata(str(query))



import warnings

print("WARNING: liquer.blueprint is deprecated, use liquer.server.blueprint")
warnings.warn(
    "liquer.blueprint is deprecated, use liquer.server.blueprint",
    DeprecationWarning,
    stacklevel=2,
)
from liquer.server.blueprint import *



from os import makedirs, name, remove
from pathlib import Path
import json
from io import BytesIO
from liquer.constants import *
import liquer.util as util
import hashlib
from liquer.metadata import Metadata
import traceback

"""Store is a flexible filesystem-like key-value store.
Stores support reading and writing of binary data and the associated metadata.
Data from the store can typically be accessed through the resource part of the query:

file/in/store.txt/-/some_command/another_command

This query is split to two parts by /-/, the first part is the resource query (file/in/store.txt),
and the second part is the standard query (some_command/another_command).
Query is interpreted as
- first read file/in/store.txt from store
- pass it as bytes to the query,
so it is equivalent to

another_command(some_command(store.get_bytes('file/in/store.txt')))

Stores can be combined and composed in various way (see e.g. mount function and method).
Various store types are supported: memory store (MemoryStore), local filesystem store (FileStore),
overlays and other. The FileSystemStore allows to mount ftp servers, zip and files via PyFilesystem2.

There is always a default main store, which can be obtained by get_store().
This is typically passed ty commands via context.store().

By convention, the folder "web" in the get_store() holds web interface and can be accessed via
/liquer/web url. 

"""

STORE = None
WEB_STORE = None

def get_store():
    global STORE
    if STORE is None:
        STORE = MountPointStore()
    return STORE


def set_store(store):
    global STORE
    STORE = store

def get_web_store():
    global WEB_STORE
    if WEB_STORE is None:
        WEB_STORE = MountPointStore()
        get_store().mount("web", WEB_STORE)
    return WEB_STORE

def mount(key, store=None):
    if store is None:
        store=key
    if type(store) is str:
        store = FileStore(store)
    get_store().mount(key, store)

def mount_folder(key, path):
    get_store().mount(key, FileStore(path))

def web_mount(key, store=None):
    if store is None:
        store=key
    if type(store) is str:
        store = FileStore(store)
    get_web_store().mount(key, store)

def web_mount_folder(key, path):
    web_mount(key, FileStore(path))

class StoreException(Exception):
    def __init__(self, message, key=None, store=None):
        self.original_message = message
        if key is not None:
            k = key if store is None else store.to_root_key(key)
            message += f":\n  key: '{k}'"
        if store is not None:
            message += f"  store: {store}"

        super().__init__(message)
        self.key = key
        self.store = store


class KeyNotFoundStoreException(StoreException):
    def __init__(self, message="Key not found in store", key=None, store=None):
        super().__init__(message=message, key=key, store=store)


class KeyNotSupportedStoreException(StoreException):
    def __init__(self, message="Key not supported in store", key=None, store=None):
        super().__init__(message=message, key=key, store=store)

class KeyRouteNotFoundStoreException(StoreException):
    def __init__(self, message="Key route not found", key=None, store=None):
        super().__init__(message=message, key=key, store=store)

class ReadOnlyStoreException(StoreException):
    def __init__(self, message="Key is read only", key=None, store=None):
        super().__init__(message=message, key=key, store=store)

class StoreMixin:
    def with_overlay(self, overlay):
        """Create an overlay over the store.
        Overlay store will be consulted first, when the key is not supported or not found in the overlay,
        the request will be passed to the current store (self).
        Note: a.with_overlay(b) is the same as b.with_fallback(a). Both methods create an OverlayStore instance.
        """
        return OverlayStore(overlay, self)

    def with_fallback(self, fallback):
        """Create a fallback for the store.
        When the key is not supported or not found in the current store (self),
        request will be passed to fallback.
        Note: a.with_overlay(b) is the same as b.with_fallback(a). Both methods create an OverlayStore instance.
        """
        return OverlayStore(self, fallback)

    def mount(self, key, store):
        """Mount a store in a mount-point"""
        return MountPointStore(self).mount(key, store)

    def read_only(self):
        """Create a read-only proxy for the store."""
        return ReadOnlyStore(self)


def key_name(key):
    if key in ("", None):
        return ""
    return str(key.split("/")[-1])

def key_name_without_extension(key):
    if key in ("", None):
        return None
    v = key_name(key).split(".")
    if len(v)>=2:
        return v[0]
    return None

def key_extension(key):
    if key in ("", None):
        return None
    v = key_name(key).split(".")
    if len(v)>=2:
        return v[-1]
    return None

def join_key(key, name):
    if key in ("", None):
        return name
    else:
        if key.endswith("/"):
            return f"{key}{name}"
        else:
            return f"{key}/{name}"

def parent_key(key):
    if key == "":
        return None
    return "/".join(key.split("/")[:-1])

class Store(StoreMixin):
    parent_store=None
    MD5_CHECKSUM=True
    def parent_key(self, key):
        "For backwards compatibility only; use parent_key function"
        return parent_key(key)

    def key_name(self, key):
        "For backwards compatibility only; use key_name function"
        return key_name(key)
    
    def join_key(self, key, name):
        "For backwards compatibility only; use join_key function"
        return join_key(key, name)

    def default_metadata(self, key, is_dir=False):
        if key is None:
            key = ""

        return dict(
            key=key,
            fileinfo=dict(name=key_name(key), is_dir=is_dir, filesystem_path=None),
        )

    def finalize_metadata(self, metadata, key, is_dir=False, data=None, update=False):
        if data is not None:
            if type(data)!=bytes:
                print(f"WARNING: Non-binary data for '{key}': type is {type(data)}")
        if key is None:
            key = ""
        metadata["key"] = key
        update = update or data is not None
        if update:
            metadata["updated"]=util.now()
        if data is not None:
            metadata["created"]=metadata["updated"]
        metadata["fileinfo"] = metadata.get("fileinfo", {})
        metadata["fileinfo"]["name"] = key_name(key)
        metadata["fileinfo"]["is_dir"] = is_dir
        metadata["fileinfo"]["filesystem_path"] = metadata["fileinfo"].get(
            "filesystem_path"
        )
            
        if data is not None:
            metadata["fileinfo"]["size"] = len(data)
            if self.MD5_CHECKSUM and type(data)==bytes:
                metadata["fileinfo"]["md5"] = hashlib.md5(data).hexdigest()

        if metadata.get("mimetype") is None:
            mimetype = mimetype_from_extension(key_extension(key))
            metadata["mimetype"]=mimetype

        if metadata.get("type_identifier") is None:
            type_identifier = type_identifier_from_extension(key_extension(key))
            metadata["type_identifier"]=type_identifier

        return Metadata(metadata).as_dict()

    def get_bytes(self, key):
        raise KeyNotFoundStoreException(key=key, store=self)

    def get_metadata(self, key):
        """Store data and metadata."""
        raise KeyNotFoundStoreException(key=key, store=self)

    def store(self, key, data, metadata):
        """Store data and metadata."""
        raise KeyNotSupportedStoreException(key=key, store=self)

    def store_metadata(self, key, metadata):
        """Store metadata only."""
        raise KeyNotSupportedStoreException(key=key, store=self)

    def remove(self, key):
        """Remove data and metadata associated with the key."""
        raise KeyNotFoundStoreException(key=key, store=self)

    def removedir(self, key):
        """Remove directory.
        The ky must be a directory.
        It depends on the underlying store whether the directory must be empty.
        """
        raise KeyNotFoundStoreException(key=key, store=self)

    def contains(self, key):
        """Returns true if store contains the key."""
        return key == ""

    def is_dir(self, key):
        """Returns true if key is a directory."""
        return key == ""

    def keys(self):
        """List or iterator of all keys"""
        return []

    def listdir(self, key):
        """Return names inside a directory specified by key.
        To get a key, names need to be joined with the key (key/name).
        If you need keys, use the listdir_keys method.  
        """
        return []

    def listdir_keys(self, key):
        """Return keys inside a directory specified by key."""
        if key in ("",None):
            return self.listdir(key)
        else:
            return [self.join_key(key, k) for k in self.listdir(key)]

    def makedir(self, key):
        "Make a directory"
        raise KeyNotSupportedStoreException(key=key, store=self)

    def openbin(self, key, mode="r", buffering=-1):
        """Return a file handle.
        This is not necessarily always well supported, but it is required to support PyFilesystem2."""
        raise KeyNotSupportedStoreException(key=key, store=self)

    def is_supported(self, key):
        """Returns true is this store supports the supplied key.
        This allows layering Stores, e.g. by with_overlay, with_fallback
        and store selectively certain data (keys) in certain stores. 
        """
        return False

    def on_data_changed(self, key):
        """Event handler called when the data is changed."""
        pass

    def on_metadata_changed(self, key):
        """Event handler called when the metadata is changed."""
        pass

    def on_removed(self, key):
        """Event handler called when the data or directory is removed."""
        pass

    def to_root_key(self, key):
        """Convert local store key to a key in a root store.
        This is can be used e.g. to convert a key valid in a mounted (child) store to
        a key of a root store. 
        The to_root_key(key) in the root_store() should point to the same object as key in self.
        """
        if self.parent_store is None:
            return key
        return self.parent_store.to_root_key(key)

    def root_store(self):
        """Get the root store.
        Root store is the highest level store in the store system.
        The to_root_key(key) in the root_store() should point to the same object as key in self.
        """
        if self.parent_store is None:
            return self
        return self.parent_store.root_store()

    def sync(self):
        pass

    def __str__(self):
        return f"Empty store"

    def __repr__(self):
        return f"Store()"


class FileStore(Store):
    METADATA = "__metadata__"

    def __init__(self, path):
        if isinstance(path, Path):
            self.path = path
        else:
            self.path = Path(path)

    def finalize_metadata(self, metadata, key, is_dir=False, data=None, update=False):
        metadata = super().finalize_metadata(
            metadata, key=key, is_dir=is_dir, data=data, update=update
        )
        metadata["fileinfo"]["filesystem_path"] = str(self.path_for_key(key).resolve())
        return Metadata(metadata).as_dict()

    def path_for_key(self, key):
        if key in (None, ""):
            return self.path
        p = self.path / key
        assert p.name != self.METADATA
        return p

    def metadata_path_for_key(self, key):
        p = self.path / key
        assert p.name != self.METADATA
        return p.parent / self.METADATA / (p.name + ".json")

    def get_bytes(self, key):
        if not self.path_for_key(key).exists():
            raise KeyNotFoundStoreException(key=key, store=self)
        with open(self.path_for_key(key), "rb") as f:
            b=f.read()
        return b

    def get_metadata(self, key):
        p = self.path_for_key(key)
        metadata = self.default_metadata(key, p.is_dir())

        if p.is_dir():
            return self.finalize_metadata(metadata, key=key, is_dir=True)
        else:
            if self.path_for_key(key).exists():
                if self.metadata_path_for_key(key).exists():
                    with open(self.metadata_path_for_key(key)) as f:
                        try:
                            metadata.update(
                                json.loads(f.read())
                            )
                        except:
                            traceback.print_exc()
                            print(f"Removing {key} due to corrupted metadata (a)")
                            self.remove(key)
                            raise KeyNotFoundStoreException(key=key, store=self)
                else:
                    m = Metadata()
                    m.status = Status.EXTERNAL
                    return self.finalize_metadata(m.as_dict(), key=key)

            else:
                if self.metadata_path_for_key(key).exists():
                    with open(self.metadata_path_for_key(key)) as f:
                        try:
                            metadata.update(
                                json.loads(f.read())
                            )
                        except:
                            traceback.print_exc()
                            print(f"Removing {key} due to corrupted metadata (b)")
                            self.remove(key)
                            raise KeyNotFoundStoreException(key=key, store=self)

                else:
                    raise KeyNotFoundStoreException(key=key, store=self)
        return self.finalize_metadata(metadata, key=key, is_dir=False)

    def store(self, key, data, metadata):
        self.path_for_key(key).parent.mkdir(parents=True, exist_ok=True)
        self.path_for_key(key).write_bytes(data)
        self.store_metadata(
            key, self.finalize_metadata(metadata, key=key, is_dir=False, data=data)
        )
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        self.metadata_path_for_key(key).parent.mkdir(parents=True, exist_ok=True)
        metadata = self.finalize_metadata(metadata, key=key, is_dir=self.is_dir(key), update=True)
        with open(self.metadata_path_for_key(key), "w") as f:
            json.dump(metadata, f)
        self.on_metadata_changed(key)

    def remove(self, key):
        try:
            self.path_for_key(key).unlink()
        except FileNotFoundError:
            pass
        try:
            self.metadata_path_for_key(key).unlink()
        except FileNotFoundError:
            pass
        
        self.on_removed(key)

    def removedir(self, key):
        (self.path_for_key(key) / self.METADATA).rmdir()
        self.path_for_key(key).rmdir()
        self.on_removed(key)

    def contains(self, key):
        if key in ("",None):
            return True
        return self.path_for_key(key).exists()

    def is_dir(self, key):
        if key in ("",None):
            return True
        return self.path_for_key(key).is_dir()

    def keys(self, parent=None):
        d = self.listdir(parent)
        if d is None:
            return []
        else:
            for k in d:
                key = k if parent is None else parent + "/" + k
                yield key
                for kk in self.keys(key):
                    yield kk

    def listdir(self, key):
        if self.is_dir(key):
            return [
                d.name
                for d in self.path_for_key(key).iterdir()
                if d.name != self.METADATA
            ]

    def makedir(self, key):
        self.path_for_key(key).mkdir(parents=True, exist_ok=True)
        (self.path_for_key(key) / self.METADATA).mkdir(parents=True, exist_ok=True)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="rb", buffering=-1):
        mode = dict(r="rb", w="wb").get(mode, mode)
        return open(self.path_for_key(key), mode)

    def is_supported(self, key):
        return True

    def __str__(self):
        return f"File store at {self.path}"

    def __repr__(self):
        return f"FileStore('{self.path}')"


class MemoryStore(Store):
    def __init__(self):
        self.directories = set()
        self.data = {}
        self.metadata = {}

    def get_bytes(self, key):
        if key not in self.data:
            raise KeyNotFoundStoreException(key=key, store=self)
        return self.data[key]

    def get_metadata(self, key):
        if key in self.metadata:
            metadata = self.metadata[key]
        else:
            if not self.is_dir(key):
                raise KeyNotFoundStoreException(key=key, store=self)

            metadata = self.default_metadata(key, is_dir=True)

        return self.finalize_metadata(metadata, key=key, is_dir=self.is_dir(key))

    def store(self, key, data, metadata):
        self.makedir(self.parent_key(key))
        self.data[key] = data
        self.metadata[key] = self.finalize_metadata(
            metadata, key=key, is_dir=False, data=data
        )
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        metadata = self.finalize_metadata(metadata, key=key, is_dir=self.is_dir(key), update=True)
        self.metadata[key] = metadata
        self.on_metadata_changed(key)

    def remove(self, key):
        try:
            self.directories.remove(key)
        except KeyError:
            pass
        try:
            del self.data[key]
        except KeyError:
            pass
        try:
            del self.metadata[key]
        except KeyError:
            pass
        self.on_removed(key)

    def removedir(self, key):
        if len(self.listdir(key)) == 0:
            try:
                self.directories.remove(key)
            except KeyError:
                pass
        self.on_removed(key)

    def contains(self, key):
        if key in ("",None):
            return True

        return key in self.directories or key in self.data or key in self.metadata

    def is_dir(self, key):
        if key in (None, ""):
            return True
        return key in self.directories

    def keys(self):
        k = set()
        k.update(self.directories)
        k.update(self.data.keys())
        k.update(self.metadata.keys())
        return sorted(k)

    def listdir(self, key):
        if key in ("", None):
            return sorted(set(k.split("/")[0] for k in self.keys() if k.split("/")[0]!=""))
        if self.is_dir(key):
            return [key_name(k) for k in self.keys() if self.parent_key(k) == key]

    def makedir(self, key):
        while key not in (None, ""):
            self.directories.add(key)
            key = self.parent_key(key)
        self.on_data_changed(key)
        self.on_metadata_changed(key)


    def openbin(self, key, mode="r", buffering=-1):
        mode = dict(r="rb", w="wb").get(mode, mode)
        if mode == "rb":
            return BytesIO(self.data[key])
        raise Exception("openbin not supported for write yet")

    def is_supported(self, key):
        return True

    def __str__(self):
        return f"Memory store"

    def __repr__(self):
        return f"MemoryStore()"

class ProxyStore(Store):
    """Proxy to another store - can be used as a basis to override certain behaviour
    """

    def __init__(self, store):
        self._store = store

    def sync(self):
        self._store.sync()

    def get_bytes(self, key):
        return self._store.get_bytes(key)

    def get_metadata(self, key):
        return self._store.get_metadata(key)

    def store(self, key, data, metadata):
        self._store.store(key, data, metadata)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        self._store.store_metadata(key, metadata)
        self.on_metadata_changed(key)

    def remove(self, key):
        self._store.remove(key)
        self.on_removed(key)

    def removedir(self, key):
        self._store.remove(key)
        self.on_removed(key)

    def contains(self, key):
        return self._store.contains(key)

    def is_dir(self, key):
        return self._store.is_dir(key)

    def keys(self):
        return self._store.keys()

    def listdir(self, key):
        return self._store.listdir(key)

    def makedir(self, key):
        self._store.makedir(key)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="r", buffering=-1):
        return self._store.openbin(key, mode, buffering)

    def is_supported(self, key):
        return self._store.is_supported(key)

    def __str__(self):
        return f"proxy of {self._store}"

    def __repr__(self):
        return f"ProxyStore({repr(self._store)})"


class ReadOnlyStore(ProxyStore):
    """Read only proxy to a store
    """

    def store(self, key, data, metadata):
        raise ReadOnlyStoreException(key=key, store=self)

    def store_metadata(self, key, metadata):
        raise ReadOnlyStoreException(key=key, store=self)

    def remove(self, key):
        raise ReadOnlyStoreException(key=key, store=self)

    def removedir(self, key):
        raise ReadOnlyStoreException(key=key, store=self)

    def makedir(self, key):
        raise ReadOnlyStoreException(key=key, store=self)

    def __str__(self):
        return f"read only proxy of {self._store}"

    def __repr__(self):
        return f"ReadOnlyStore({repr(self._store)})"

class OverlayStore(Store):
    """Overlay store combines two stores: overlay and fallback.
    Overlay is used as a primary store for reading and writing.
    The fallback is used only for reading if key is not found in the overlay store (and is not removed).
    Thus the fallback store is never modified.
    """

    def __init__(self, overlay, fallback):
        self.overlay = overlay
        self.fallback = fallback
        self.removed = set()

    def sync(self):
        self.overlay.sync()
        self.fallback.sync()

    def get_bytes(self, key):
        if key not in self.removed:
            if self.overlay.contains(key):
                return self.overlay.get_bytes(key)
            else:
                return self.fallback.get_bytes(key)

    def get_metadata(self, key):
        if key not in self.removed:
            if self.overlay.contains(key):
                return self.overlay.get_metadata(key)
            else:
                return self.fallback.get_metadata(key)
        raise KeyNotFoundStoreException(key=key, store=self)

    def store(self, key, data, metadata):
        try:
            self.removed.remove(key)
        except KeyError:
            pass
        self.overlay.store(key, data, metadata)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        try:
            self.removed.remove(key)
        except KeyError:
            pass
        self.overlay.store_metadata(key, metadata)
        self.on_metadata_changed(key)

    def remove(self, key):
        if key not in self.removed:
            if self.overlay.contains(key):
                self.overlay.remove(key)
            if self.fallback.contains(key):
                self.removed.add(key)
        self.on_removed(key)

    def removedir(self, key):
        if self.contains(key):
            if len(self.listdir(key)) == 0:
                if self.overlay.contains(key):
                    self.overlay.remove(key)
                else:
                    self.removed.add(key)
        self.on_removed(key)

    def contains(self, key):
        if key in self.removed:
            return False
        else:
            return self.overlay.contains(key) or self.fallback.contains(key)

    def is_dir(self, key):
        if key in self.removed:
            return False
        else:
            if self.overlay.contains(key):
                return self.overlay.is_dir(key)
            else:
                return self.fallback.is_dir(key)

    def keys(self):
        return sorted(
            set(self.overlay.keys())
            .union(self.fallback.keys())
            .difference(self.removed)
        )

    def listdir(self, key):
        if key.endswith("/"):
            key = key[:-1]
        ld = self.overlay.listdir(key)
        d = set() if ld is None else set(ld)
        ld = self.fallback.listdir(key)
        d = d if ld is None else d.union(ld)
        return [x for x in sorted(d) if key + "/" + x not in self.removed]

    def makedir(self, key):
        if key in self.removed:
            self.removed.remove(key)
        self.overlay.makedir(key)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="r", buffering=-1):
        mode = dict(r="rb", w="wb").get(mode, mode)
        if mode == "rb":
            if self.overlay.contains(key):
                return self.overlay.openbin(key, mode, buffering)
            elif self.fallback.contains(key):
                return self.fallback.openbin(key, mode, buffering)
        elif mode == "wb":
            return self.overlay.openbin(key, mode, buffering)

    def is_supported(self, key):
        return self.overlay.is_supported(key) or self.fallback.is_supported(key())

    def __str__(self):
        return f"Overlay of ({self.overlay}) over ({self.fallback})"

    def __repr__(self):
        return f"OverlayStore({repr(self.overlay)},{repr(self.fallback)})"


class RoutingStore(Store):
    def route_to(self, key):
        raise KeyRouteNotFoundStoreException(key=key, store=self)

    def finalize_metadata(self, metadata, key, is_dir=False, data=None, update=False):
        metadata = super().finalize_metadata(
            metadata, key=key, is_dir=is_dir, data=data, update=update
        )
        return metadata

    def is_supported(self, key):
        try:
            store = self.route_to(key)
            if store is None:
                return False
            return store.is_supported(key)
        except KeyNotSupportedStoreException:
            return False

    def get_bytes(self, key):
        return self.route_to(key).get_bytes(key)

    def get_metadata(self, key):
        metadata=self.route_to(key).get_metadata(key)
        metadata["key"]=key
        return metadata

    def store(self, key, data, metadata):
        self.route_to(key).store(key, data, metadata)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        self.route_to(key).store_metadata(key, metadata)
        self.on_metadata_changed(key)

    def remove(self, key):
        self.route_to(key).remove(key)
        self.on_removed(key)

    def removedir(self, key):
        self.route_to(key).removedir(key)
        self.on_removed(key)

    def contains(self, key):
        return self.route_to(key).contains(key)

    def is_dir(self, key):
        return self.route_to(key).is_dir(key)

    def keys(self):
        raise NotImplementedError

    def listdir(self, key):
        return self.route_to(key).listdir(key)

    def makedir(self, key):
        self.route_to(key).makedir(key)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="r", buffering=-1):
        return self.route_to(key).openbin(key, mode=mode, buffering=buffering)

    def __str__(self):
        return f"Abstract routing store"

    def __repr__(self):
        return f"RoutingStore()"


class KeyTranslatingStore(Store):
    def __init__(self, store):
        self.substore = store
        self.substore.parent_store=self

    def sync(self):
        self.substore.sync()

    def translate_key(self, key, inverse=False):
        return key

    def to_root_key(self, key):
        """Convert local store key to a key in a root store.
        This is can be used e.g. to convert a key valid in a mounted (child) store to
        a key of a root store. 
        """
        key = self.translate_key(key, inverse=True)
        if self.parent_store is None:
            return key
        return self.parent_store.to_root_key(key)

    def is_supported(self, key):
        try:
            tkey = self.translate_key(key)
            if tkey is None:
                return False
            return self.substore.is_supported(tkey)
        except KeyNotSupportedStoreException:
            return False

    def get_bytes(self, key):
        return self.substore.get_bytes(self.translate_key(key))

    def get_metadata(self, key):
        tkey = self.translate_key(key)
        metadata=self.substore.get_metadata(tkey)
        metadata["key"]= key
        if "recipes_key" in metadata:
            metadata["recipes_key"] = self.translate_key(metadata["recipes_key"], inverse=True)
        return metadata

    def store(self, key, data, metadata):
        self.substore.store(self.translate_key(key), data, metadata)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        self.substore.store_metadata(self.translate_key(key), metadata)
        self.on_metadata_changed(key)

    def remove(self, key):
        self.substore.remove(self.translate_key(key))
        self.on_removed(key)

    def removedir(self, key):
        self.substore.removedir(self.translate_key(key))
        self.on_removed(key)

    def contains(self, key):
        return self.substore.contains(self.translate_key(key))

    def is_dir(self, key):
        return self.substore.is_dir(self.translate_key(key))

    def keys(self):
        for key in self.substore.keys():
            yield self.translate_key(key, inverse=True)

    def listdir(self, key):
        return self.substore.listdir(self.translate_key(key))

    def makedir(self, key):
        self.substore.makedir(self.translate_key(key))
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="r", buffering=-1):
        return self.substore.openbin(
            self.translate_key(key), mode=mode, buffering=buffering
        )

    def __str__(self):
        return f"Key translating store on ({self.substore})"

    def __repr__(self):
        return f"KeyTranslatingStore({repr(self.substore)})"


class PrefixStore(KeyTranslatingStore):
    def __init__(self, store, prefix):
        self.substore = store
        self.prefix = prefix
        self.substore.parent_store=self

    def translate_key(self, key, inverse=False):
        prefix = self.prefix + "/"
        if inverse:
            if key in (None, ""):
                return self.prefix
            else:
                return prefix + key
        else:
            if key == self.prefix:
                return ""
            else:
                if key.startswith(prefix):
                    return key[len(prefix) :]
                else:
                    raise KeyNotSupportedStoreException

    def contains(self, key):
        if key == self.prefix:
            return True
        return self.substore.contains(self.translate_key(key))

    def is_dir(self, key):
        if key == self.prefix:
            return True
        return self.substore.is_dir(self.translate_key(key))

    def __str__(self):
        return f"Prefixed store {self.prefix} on ({self.substore})"

    def __repr__(self):
        return f"PrefixStore({repr(self.substore)}, prefix={repr(self.prefix)})"


class MountPointStore(RoutingStore):
    def __init__(self, default_store=None, routing_table=None):
        self.default_store = default_store
        self.routing_table = [] if routing_table is None else routing_table
    def sync(self):
        for key, store in self.routing_table:
            print("SYNC ",key)
            store.sync()
        if self.default_store is not None:
            self.default_store.sync()

    def umount(self, umount_key):
        self.routing_table = [(key, store) for key, store in self.routing_table if key!=umount_key]
        return self

    def mount(self, key, store):
        self.umount(key)
        self.routing_table.append((key, PrefixStore(store, prefix=key)))
        return self

    def route_to(self, key):
        for prefix, store in reversed(self.routing_table):
            if key == prefix:
                return store
            if not prefix.endswith("/"):
                prefix += "/"
            if key.startswith(prefix):
                if store.is_supported(key):
                    return store
        if self.default_store is not None:
            return self.default_store
        raise KeyRouteNotFoundStoreException(key=key, store=self)

    def get_metadata(self, key):
        try:
            metadata=self.route_to(key).get_metadata(key)
            metadata["key"]=key
            
            return metadata
        except KeyRouteNotFoundStoreException:
            if self.is_dir(key):
                return self.finalize_metadata({}, key, is_dir=True)
        raise KeyNotFoundStoreException(key=key, store=self)

    def is_dir(self, key):
        if key=="":
            return True
        try:
            return self.route_to(key).is_dir(key)
        except KeyRouteNotFoundStoreException:
            for route, _ in reversed(self.routing_table):
                if route==key or route.startswith(key+"/"):
                    return self.finalize_metadata({}, key, is_dir=True)
        return False

    def keys(self):
        prefixes = []
        for prefix, store in reversed(self.routing_table):
            yield prefix
            if not prefix.endswith("/"):
                prefix += "/"
            for key in store.keys():
                if any(key.startswith(p) for p in prefixes):
                    continue
                if key.startswith(prefix):
                    yield key
            prefixes.append(prefix)
        if self.default_store is not None:
            for key in self.default_store.keys():
                if any(key.startswith(p) for p in prefixes):
                    continue
                yield key

    def listdir(self, key):
        try:
            store = self.route_to(key)
        except KeyRouteNotFoundStoreException:
            store = None

        if store is None:
            d = set()
        else:
            d = set(store.listdir(key) or [])
        key_split = key.split("/")
        if len(key_split) == 1 and key_split[0] == "":
            key_split = []
        key_depth = len(key_split)
        for prefix, _ in self.routing_table:
            if prefix.startswith(key + "/") or key in (None, ""):
                v = prefix.split("/")
                d.add(v[key_depth])
        if self.default_store is not None:
            for prefix in self.default_store.keys():
                if prefix.startswith(key + "/") or key in (None, ""):
                    v = prefix.split("/")
                    d.add(v[key_depth])

        return sorted(d)

    def removedir(self, key):
        for k, store in self.routing_table:
            if k == key:
                raise StoreException(
                    f"Can't remove {key} because it is a mount point of {repr(store)}",
                    key=key,
                    store=self,
                )

        return self.route_to(key).removedir(key)

    def __str__(self):
        return f"Mount point store routed by {self.routing_table} with default store {self.default_store}"

    def __repr__(self):
        return f"MountPointStore({repr(self.default_store)}, routing_table={repr(self.routing_table)})"


class FileSystemStore(Store):
    METADATA = "__metadata__"

    def __init__(self, fs, path=""):
        if isinstance(path, Path):
            self.path = path
        else:
            self.path = Path(path)
        self.fs = fs

    def path_for_key(self, key):
        if key in (None, ""):
            return str(self.path)
        p = self.path / key
        assert p.name != self.METADATA
        return str(p)

    def metadata_path_for_key(self, key):
        p = self.path / key
        assert p.name != self.METADATA
        return str(p.parent / self.METADATA / (p.name + ".json"))

    def metadata_dir_path_for_key(self, key):
        p = self.path / key
        assert p.name != self.METADATA
        return str(p.parent / self.METADATA)

    def get_bytes(self, key):
        if not self.fs.exists(self.path_for_key(key)):
            raise KeyNotFoundStoreException(
                f"Can't find {self.path_for_key(key)} in filesystem {self.fs}",
                key=key,
                store=self,
            )
        return self.fs.readbytes(self.path_for_key(key))

    def get_metadata(self, key):
        p = self.path_for_key(key)
        isdir = self.fs.isdir(p)
        metadata = self.default_metadata(key, isdir)
        if isdir:
            return self.finalize_metadata(metadata, key, is_dir=True)
        else:
            if self.fs.exists(self.path_for_key(key)):
                if self.fs.exists(self.metadata_path_for_key(key)):
                    try:
                        metadata.update(
                            json.loads(self.fs.readtext(self.metadata_path_for_key(key)))
                        )
                    except:
                        traceback.print_exc()
                        print(f"Removing {key} due to corrupted metadata (c)")
                        self.remove(key)
                        raise KeyNotFoundStoreException(key=key, store=self)

            else:
                if self.fs.exists(self.metadata_path_for_key(key)):
                    try:
                        metadata.update(
                            json.loads(self.fs.readtext(self.metadata_path_for_key(key)))
                        )
                    except:
                        traceback.print_exc()
                        print(f"Removing {key} due to corrupted metadata (d)")
                        self.remove(key)
                        raise KeyNotFoundStoreException(key=key, store=self)
                else:
                    raise KeyNotFoundStoreException(key=key, store=self)
            return self.finalize_metadata(metadata, key, is_dir=False)

    def store(self, key, data, metadata):
        self.fs.makedirs(self.path_for_key(self.parent_key(key)))
        self.fs.writebytes(self.path_for_key(key), data)
        assert self.fs.exists(self.path_for_key(key))
        self.store_metadata(
            key, self.finalize_metadata(metadata, key=key, is_dir=False, data=data)
        )
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        parent = self.metadata_dir_path_for_key(key)
        metadata = self.finalize_metadata(metadata, key=key, is_dir=self.is_dir(key), update=True)
        self.fs.makedirs(parent)
        with self.fs.open(self.metadata_path_for_key(key), "w") as f:
            json.dump(metadata, f)
        self.on_metadata_changed(key)

    def remove(self, key):
        if not self.contains(key):
            raise KeyNotFoundStoreException(key=key, store=self)
        if self.is_dir(key):
            raise KeyNotFoundStoreException(
                message=f"Can't remove a directory {key}; use removedir",
                key=key,
                store=self,
            )
        try:
            self.fs.remove(self.path_for_key(key))
        except:
            pass
        try:
            self.fs.remove(self.metadata_path_for_key(key))
        except:
            pass
        self.on_removed(key)

    def removedir(self, key):
        metadir = self.path_for_key(key) + "/" + self.METADATA
        try:
            self.fs.removedir(metadir)
        except:
            pass
        self.fs.removedir(self.path_for_key(key))
        self.on_removed(key)

    def contains(self, key):
        if key in ("",None):
            return True

        return self.fs.exists(self.path_for_key(key))

    def is_dir(self, key):
        if key in ("",None):
            return True
        return self.fs.isdir(self.path_for_key(key))

    def keys(self, parent=None):
        d = self.listdir(parent)
        if d is None:
            return []
        else:
            for k in d:
                key = k if parent is None else parent + "/" + k
                yield key
                for kk in self.keys(key):
                    yield kk

    def listdir(self, key):
        if self.is_dir(key):
            return [
                key_name(d)
                for d in self.fs.listdir(self.path_for_key(key))
                if key_name(d) != self.METADATA
            ]

    def makedir(self, key):
        self.fs.makedir(self.path_for_key(key), recreate=True)
        self.fs.makedir(self.path_for_key(key) + "/" + self.METADATA, recreate=True)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="rb", buffering=-1):
        return self.fs.openbin(self.path_for_key(key), mode=mode, buffering=buffering)

    def is_supported(self, key):
        return True

    def __str__(self):
        return f"Filesystem {self.fs} store at {self.path}"

    def __repr__(self):
        return f"FileSystemStore({repr(self.fs)}, {repr(self.path)})"



from io import StringIO, BytesIO
from urllib.request import urlopen
import pandas as pd
import hxl
from liquer.state_types import StateType, register_state_type
from liquer.commands import command, first_command
from liquer.constants import mimetype_from_extension


class HxlStateType(StateType):
    def identifier(self):
        "Define an unique string identifier for the state type"
        return "hxl_dataset"

    def default_extension(self):
        "Default file extension for the state type"
        return "csv"

    def is_type_of(self, data):
        "Check if data is of this state type"
        return isinstance(data, hxl.model.Dataset)

    def as_bytes(self, data, extension=None):
        """Serialize data as bytes
        File extension may be provided and influence the serialization format.
        """
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)
        if extension == "csv":
            output = "".join(data.gen_csv(show_headers=True, show_tags=True))
            return output.encode("utf-8"), mimetype
        elif extension == "json":
            output = "".join(data.gen_json(show_headers=True, show_tags=True))
            return output.encode("utf-8"), mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by HXL dataset type."
            )

    def from_bytes(self, b: bytes, extension=None):
        """De-serialize data from bytes
        File extension may be provided and influence the serialization format.
        """
        if extension is None:
            extension = self.default_extension()
        f = BytesIO()
        f.write(b)
        f.seek(0)

        if extension == "csv":
            return hxl.data(f)
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by HXL dataset type."
        )

    def copy(self, data):
        """Make a deep copy of the data"""
        return data

    def data_characteristics(self, data):
        return dict(description=f"HXL dataset")

HXL_DATASET_STATE_TYPE = HxlStateType()
register_state_type(hxl.Dataset, HXL_DATASET_STATE_TYPE)
register_state_type(hxl.io.HXLReader, HXL_DATASET_STATE_TYPE)


@first_command
def hxl_from(url):
    """Load data from URL"""
    return hxl.data(url)


@command
def hxl2df(data):
    """Convert hxl dataset to pandas dataframe"""
    f = BytesIO()
    for line in data.gen_csv(show_headers=True, show_tags=True):
        f.write(line.encode("utf-8"))
    f.seek(0)
    return pd.read_csv(f)


@command
def df2hxl(df):
    """Convert pandas dataframe to hxl dataset"""
    f = StringIO()
    df.to_csv(f, index=False)
    f = BytesIO(f.getvalue().encode("utf-8"))
    return hxl.data(f)


@command
def set_all_tags(df, *tags):
    """Set all tags in pandas dataframe. All tags should be specified in the same order as columns.
    If not all tags are specified, empty strings are appended.
    Creates a first row with tags in a dataframe.
    Hash characters are automatically prepended if not present.
    """

    tags = pd.DataFrame(
        [
            {
                c: tag.strip() if tag.strip().startswith("#") else "#" + tag.strip()
                for c, tag in zip(df.columns, tags)
            }
        ],
        columns=df.columns,
    )
    tags.fillna(value="", inplace=True)
    return tags.append(df, ignore_index=True)


@command
def set_tags(df, *column_tag):
    """Set tags in pandas dataframe, tags are specified as a list with alternating column, tag.
    Creates a first row with tags in a dataframe.
    Hash characters are automatically prepended if not present.
    """
    if len(column_tag) % 2 != 0:
        raise Exception(
            f"Columns-tags (len:{len(column_tag)})argument od set_tags do not form pairs {column_tag}"
        )
    tags = pd.DataFrame(
        [
            {
                column_tag[i]: column_tag[i + 1].strip()
                if column_tag[i + 1].strip().startswith("#")
                else "#" + column_tag[i + 1].strip()
                for i in range(0, len(column_tag), 2)
            }
        ],
        columns=df.columns,
    )
    tags.fillna(value="", inplace=True)
    return tags.append(df, ignore_index=True)



from liquer.commands import command, first_command, command_registry
import liquer.ext.basic
import inspect


@first_command(ns="py")
def module_source(module_name):
    "Return source code for the module"
    assert all((c.isalnum() or (c in "._")) for c in module_name)
    exec(f"import {module_name}")
    mod = eval(module_name)
    return inspect.getsource(mod)



from copy import deepcopy

class DependencyException(Exception):
    def __init__(self, message, query=None):
        self.original_message = message
        if query is not None:
            message += f":\n  query: '{query}'"

        super().__init__(message)
        self.query = query

class VersionCollisionException(Exception):
    def __init__(self, message, query=None):
        self.original_message = message
        if query is not None:
            message += f":\n  query: '{query}'"

        super().__init__(message)
        self.query = query

class CommandVersionCollisionException(VersionCollisionException):
    def __init__(self, message=None, query=None, command=None, ns=None):
        if message is None:
            message = f"Version collision for command '{command}' in namespace '{ns}' for query '{query}'"
        super().__init__(message=message, query=query)


class Dependencies:
    def __init__(self, dependencies=None):
        if dependencies is None:
            dependencies={}
        self.set_dependencies(dependencies)

    def set_dependencies(self, dependencies):
        if "query" not in dependencies:
            dependencies["query"] = ""
        if "commands" not in dependencies:
            dependencies["commands"] = {}
        self.dependencies = dependencies
        return self

    def as_dict(self):
        return deepcopy(self.dependencies)

    @property
    def query(self):
        return self.dependencies["query"]

    @query.setter
    def query(self, value):
        self.dependencies["query"] = value

    def add_command_dependency(self, ns, command_metadata, detect_collisions=True):
        key = f"ns-{ns}/{command_metadata.name}"
        version = command_metadata.version
        if detect_collisions:
            if key in self.dependencies["commands"]:
                old_version=self.dependencies["commands"][key]
                if old_version!=version:
                    print(f"Version collision for command '{command_metadata.name}' in namespace '{ns}' for query '{self.query}'")
                    print(f"Existing version: {old_version}")
                    print(f"New version:      {version}")
                    raise CommandVersionCollisionException(command=command_metadata.name, ns=ns, query=self.query)
        self.dependencies["commands"][key]=version
        return self


from io import StringIO, BytesIO
import pandas as pd
from liquer.state_types import StateType, register_state_type
from liquer.constants import mimetype_from_extension

from liquer.commands import command, first_command
from openpyxl import Workbook, load_workbook
from openpyxl.utils.dataframe import dataframe_to_rows
from tempfile import TemporaryDirectory
from pathlib import Path
from liquer.context import get_context
from itertools import islice

class OpenpyxlWorkbookStateType(StateType):
    def identifier(self):
        return "openpyxl_workbook"

    def default_extension(self):
        return "xlsx"

    def is_type_of(self, data):
        return isinstance(data, Workbook)

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)

        if extension in ("xlsx","xltx"):
            with TemporaryDirectory() as tmpdir:
                path = Path(tmpdir) / f"data.{extension}"
                data.save(str(path))
                b=path.read_bytes()
                return b, mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by openpyxl_workbook type."
            )

    def from_bytes(self, b: bytes, extension=None):
        if extension is None:
            extension = self.default_extension()
        f = BytesIO()
        f.write(b)
        f.seek(0)
        if extension in ("xlsx","xltx"):
            return load_workbook(f)
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by openpyxl_workbook type."
        )

    def copy(self, data):
        return self.from_bytes(self.as_bytes(data)[0])

    def data_characteristics(self, data):
        return dict(description=f"Excel workbook with {len(data.sheetnames)} sheets.")
        

OPENPYXL_WORKBOOK_STATE_TYPE = OpenpyxlWorkbookStateType()
register_state_type(Workbook, OPENPYXL_WORKBOOK_STATE_TYPE)

@command
def workbook(data, index=True, header=True, context=None):
    """Convert bytes or a dataframe to a workbook"""
    context=get_context(context)
    if type(data)==bytes:
        context.info("Workbook from bytes")
        return OPENPYXL_WORKBOOK_STATE_TYPE.from_bytes(data)
    elif isinstance(data,pd.DataFrame):
        context.info("Workbook from pandas DataFrame")
        wb=Workbook()
        ws=wb.active
        for r in dataframe_to_rows(df, index=index, header=header):
            ws.append(r)
        return wb
    elif isinstance(data,Workbook):
        return data
    raise Exception(f"Unsupported workbook type: {type(data)}")

@command
def workbook_sheet_df(wb, sheet=None, context=None):
    """Extract a workbook sheet as a data-frame"""
    context = get_context(context)
    if type(wb)==bytes:
        wb = workbook(wb, context=context)
    if sheet in ("",None):
        context.info("Using active sheet")
        ws=wb.active
    else:
        ws=wb[sheet]
    try:
        i=int(sheet)
        sheet = wb.sheetnames[i]
        context.info(f"Using sheet {i} with name '{sheet}'")
    except:
        pass

    data = ws.values
    cols = next(data)[1:]
    data = list(data)
    idx = [r[0] for r in data]
    data = (islice(r, 1, None) for r in data)
    df = pd.DataFrame(data, index=idx, columns=cols)    
    return df


import traceback
from collections import namedtuple
import inspect
from liquer.state import State
from liquer.parser import encode
import marshal
import pickle
import types
import traceback
import base64
from liquer.parser import (
    StringActionParameter,
    ActionParameter,
    QueryException,
    ExpandedActionParameter,
)


"""This module is responsible for registering commands
Commands are composed of a command executable and command metadata, which are collected in a command registry.
CommandRegistry is a singleton that can be obtained by get_command_registry().

Command metadata (CommandMetadata tuple) contain informations about the command and its arguments
(type, parsing and editing of each argument). These metadata are a basis for parsing of arguments as well as command editor creation.

Argument parsers are responsible for parsing command arguments into desired types.
ArgumentParser has a parse method, which takes argument metadata and list of arguments (typically supplied as list of strings
resulting from liquer.parser.decode). Parsing may extract arbitrary amount of arguments and thus support more complex data structures.
Parse method returns a tuple with the parsed argument and remaining unparsed arguments.
Argument parsers which do not need multiple instances (typical case) have predefined constants.
Multiple argument parsers may be collected in a SequenceArgumentParser by use of + operator (e.g. INT_AP + FLOAT_AP).

Though commands can be registered with a low level method 'register' of CommandRegistry (which allows the greatest flexibility),
the "mainstream" way of command registration is by simply decorating a function with @command or @first_command.

"""

CommandMetadata = namedtuple(
    "CommandMetadata",
    ["name", "label", "module", "doc", "state_argument", "arguments", "attributes", "version"],
)


_remote_registration = False


def is_remote_registration_enabled():
    """Returns true if remote registration is enabled.
    This is a flag that a web interface can use to enable or disable the remote command registration.

    WARNING: Remote command registration allows to deploy arbitrary python code on LiQuer server,
    therefore it is a HUGE SECURITY RISK and it only should be used if other security measures are taken
    (e.g. on localhost or intranet where only trusted users have access).
    This is on by default on Jupyter server extension.
    """
    return _remote_registration


def enable_remote_registration():
    """Enable remote registration service

    WARNING: Remote command registration allows to deploy arbitrary python code on LiQuer server,
    therefore it is a HUGE SECURITY RISK and it only should be used if other security measures are taken
    (e.g. on localhost or intranet where only trusted users have access).
    This is on by default on Jupyter server extension.
    """
    global _remote_registration
    _remote_registration = True


def disable_remote_registration():
    "Disable remote registration service"
    global _remote_registration
    _remote_registration = True


class RegisterRemoteMixin:
    def register_remote_serialized(self, b):
        """Helper method used to register serialized command in remote registration"""
        if is_remote_registration_enabled():
            try:
                f, metadata, modify = self.decode_registration(b)
                self.register_command(f, metadata, modify=modify)
                ns = metadata.attributes.get("ns", "root")
                return dict(
                    message=f"Function {f.__name__} in namespace {ns} is registered as command",
                    status="OK",
                )
            except:
                return dict(
                    message="Error while registering command",
                    traceback=traceback.format_exc(),
                    status="ERROR",
                )
        else:
            return dict(
                message="Remote command registration is disabled.", status="ERROR"
            )

    @classmethod
    def encode_registration(cls, f, metadata, modify=False):
        code = marshal.dumps(f.__code__)
        return b"B" + pickle.dumps(
            (
                code,
                f.__name__,
                f.__defaults__,
                f.__closure__,
                metadata._asdict(),
                modify,
            )
        )

    @classmethod
    def encode_registration_base64(cls, f, metadata, modify=False):
        return b"E" + base64.urlsafe_b64encode(
            cls.encode_registration(f, metadata, modify)
        )

    @classmethod
    def decode_registration(cls, b):
        assert type(b) == bytes
        if b[0] == b"E"[0]:
            print("DECODE E:", b[:20])
            print()
            b = base64.urlsafe_b64decode(b[1:])
        print("DECODE B:", b[:20])
        print()
        print()
        assert b[0] == b"B"[0]
        b = b[1:]

        code, name, defaults, closure, metadata, modify = pickle.loads(b)
        bc = marshal.loads(code)
        f = types.FunctionType(bc, globals(), name, defaults, closure)
        return (f, CommandMetadata(**metadata), modify)


class RemoteCommandRegistry(RegisterRemoteMixin, object):
    """Remote command registry allows to register commands into a remote LiQuer server, e.g. Jupyter server extension"""

    def __init__(self, url, use_get_method=False):
        self.url = url
        self.use_get_method = use_get_method

    def register_command(self, f, metadata, modify=False):
        import requests

        url = self.url
        url += "" if url[-1] == "/" else "/"

        if self.use_get_method:
            b = self.encode_registration_base64(f, metadata, modify)
            encoded = b.decode("ascii")
            response = requests.get(url=url + encoded)
        else:
            b = self.encode_registration(f, metadata, modify)
            response = requests.post(url=url, data=b)

        if response.ok:
            response = response.json()
            if response["status"] != "OK":
                print(response.get("traceback", ""))
                raise Exception(
                    "Remote registration failed: "
                    + response.get("message", f"Error registering {f.__name__}")
                )
        else:
            response.raise_for_status()


class CommandRegistry(RegisterRemoteMixin, object):
    """Class responsible for registering all commands and their metadata"""

    def __init__(self):
        """Create empty command registry"""
        self.executables = {}
        self.metadata = {}
        self.namespaces = {}

    def is_doubleregistered(self, executable, metadata):
        """Returns True if the same function is already registered as a command.
        Another function registered under the same name would return False.
        """
        name = metadata.name
        ns = metadata.attributes.get("ns", "root")
        self.executables[ns] = self.executables.get(ns, {})
        self.metadata[ns] = self.metadata.get(ns, {})

        if name in self.executables[ns]:
            registered = self.metadata[ns][name]
            return (
                name == registered.name
                and executable.inner_id() == self.executables[ns][name].inner_id()
            )
        return False

    def register_command(self, f, metadata, modify=False):
        parser = argument_parser_from_command_metadata(metadata)
        if metadata.state_argument is None:
            executable = FirstCommandExecutable(f, metadata, parser)
        else:
            executable = CommandExecutable(f, metadata, parser)
        return self.register(executable, metadata)

    def register(self, executable, metadata, modify=False):
        """Create command
        executable is an CommandExecutable of the command,
        metadata is CommandMetadata
        """
        name = metadata.name
        ns = metadata.attributes.get("ns", "root")

        modify = modify or metadata.attributes.get("modify_command", False)
        can_register = modify
        self.executables[ns] = self.executables.get(ns, {})
        self.metadata[ns] = self.metadata.get(ns, {})
        if name in self.executables[ns]:
            if self.is_doubleregistered(executable, metadata):
                can_register = True
        else:
            can_register = True
        if can_register:
            self.executables[ns][name] = executable
            self.metadata[ns][name] = metadata
        else:
            raise Exception(f"Command {name} is already registered")

    def as_dict(self):
        """Returns dictionary representation of the registry, safe to serialize as json"""
        return {
            ns: {name: cmd._asdict() for name, cmd in metadata.items()}
            for ns, metadata in self.metadata.items()
        }

    def evaluate_command_old(self, state, qcommand: list):
        if not state.is_volatile():
            state = state.clone()
        command_name = qcommand[0]
        ns, command, metadata = self.resolve_command(state, command_name)
        if command is None:
            print(f"Unknown command: {command_name}")
            return state.with_data(None).log_error(
                message=f"Unknown command: {command_name}"
            )
        else:
            try:
                state = command(state, *qcommand[1:])
            except Exception as e:
                traceback.print_exc()
                state.log_exception(message=str(e), traceback=traceback.format_exc())
                state.exception = e
        arguments = getattr(state, "arguments", None)
        state.metadata["commands"].append(qcommand)
        state.metadata["extended_commands"].append(
            dict(
                command_name=command_name,
                ns=ns,
                qcommand=qcommand,
                command_metadata=metadata._asdict(),
                arguments=arguments,
            )
        )
        state.query = encode(state.metadata["commands"])
        state.metadata["attributes"] = {
            key: value
            for key, value in state.metadata["attributes"].items()
            if key[0].isupper()
        }
        if metadata is not None:
            state.metadata["attributes"].update(metadata.attributes)

        return state

    def resolve_command(self, state, command_name):
        for ns in state.vars.get("active_namespaces", ["root"]):
            if ns not in self.executables:
                print(f"Unknown namespace: {ns}")
                continue
            if command_name in self.executables[ns]:
                break

        if command_name in self.executables[ns]:
            command = self.executables[ns][command_name]
            if command_name in self.metadata[ns]:
                return ns, command, self.metadata[ns][command_name]
            else:
                print(f"Unknown command (metadata): {command_name}")
                return ns, command, None

        return None, None, None


_command_registry = None


def command_registry():
    """Return global the command registry object"""
    global _command_registry
    if _command_registry is None:
        _command_registry = CommandRegistry()
    return _command_registry


def remote_command_registry(url, use_get_method=False):
    """Configure the command registry to use remote LiQuer server
    Provide url to the registration service
    """
    global _command_registry
    _command_registry = RemoteCommandRegistry(url, use_get_method=use_get_method)
    return _command_registry


def reset_command_registry():
    """Create empty global command registry"""
    global _command_registry
    _command_registry = CommandRegistry()
    return _command_registry


class ArgumentParserException(QueryException):
    def __init__(self, message, position=None, query=None):
        super().__init__(message, position, query)


class ArgumentParser(object):
    is_argv = False

    def __add__(self, ap):
        return SequenceArgumentParser(self, ap)

    def parse(self, metadata, args):
        return args[0], args[1:]

    def parse_meta(self, metadata, args, context=None):
        query = None if context is None else context.raw_query
        if isinstance(args[0], str):
            return args[0], (args[0], metadata), args[1:]
        elif isinstance(args[0], StringActionParameter):
            return args[0].string, (args[0].string, metadata), args[1:]
        elif isinstance(args[0], ExpandedActionParameter):
            return args[0].value, (args[0].value, metadata), args[1:]
        elif isinstance(args[0], ActionParameter):
            raise ArgumentParserException(
                f"Unsupported action parameter {repr(args[0])}  in ArgumentParser",
                position=args[0].position,
                query=query,
            )
        else:
            return args[0], (args[0], metadata), args[1:]


class SequenceArgumentParser(ArgumentParser):
    def __init__(self, *sequence):
        self.sequence = list(sequence)

    def __add__(self, ap):
        return SequenceArgumentParser(*(self.sequence + [ap]))

    def __iadd__(self, ap):
        self.sequence.append(ap)
        return self

    def parse(self, metadata, args):
        parsed_arguments = []
        for ap, meta in zip(self.sequence, metadata):
            parsed, args = ap.parse(meta, args)
            if ap.is_argv:
                parsed_arguments.extend(parsed)
            else:
                parsed_arguments.append(parsed)
        return parsed_arguments, args

    def parse_meta(self, metadata, args, context=None):
        parsed_arguments = []
        parsed_meta = []
        for ap, meta in zip(self.sequence, metadata):
            parsed, argmeta, args = ap.parse_meta(meta, args, context=context)
            if ap.is_argv:
                parsed_arguments.extend(parsed)
            else:
                parsed_arguments.append(parsed)
            parsed_meta.append(argmeta)
        return parsed_arguments, parsed_meta, args


class ContextArgumentParser(ArgumentParser):
    def parse(self, metadata, args):
        return None, args

    def parse_meta(self, metadata, args, context=None):
        return context, (context, metadata), args


class IntArgumentParser(ArgumentParser):
    def parse(self, metadata, args):
        return int(args[0]), args[1:]

    def parse_meta(self, metadata, args, context=None):
        query = None if context is None else context.raw_query
        if isinstance(args[0], str):
            try:
                value = int(args[0])
            except:
                raise ArgumentParserException(
                    f"""Error parsing integer argument '{metadata["name"]}' from string {repr(args[0])}""",
                    query=query,
                )
        elif isinstance(args[0], StringActionParameter):
            try:
                value = int(args[0].string)
            except:
                raise ArgumentParserException(
                    f"""Error parsing integer argument '{metadata["name"]}' from {repr(args[0].string)}""",
                    position=args[0].position,
                    query=query,
                )
        elif isinstance(args[0], ExpandedActionParameter):
            try:
                value = int(args[0].value)
            except:
                raise ArgumentParserException(
                    f"""Error parsing integer argument '{metadata["name"]}' from link {repr(args[0].link.encode())}""",
                    position=args[0].position,
                    query=query,
                )
        elif isinstance(args[0], ActionParameter):
            raise ArgumentParserException(
                f"Unsupported action parameter type {repr(args[0])} in IntArgumentParser",
                position=args[0].position,
                query=query,
            )
        else:
            try:
                value = int(args[0])
            except:
                raise ArgumentParserException(
                    f"""Error parsing integer argument '{metadata["name"]}' from object {repr(args[0])}""",
                    query=query,
                )

        return value, (value, metadata), args[1:]


class FloatArgumentParser(ArgumentParser):
    def parse(self, metadata, args):
        return float(args[0]), args[1:]

    def parse_meta(self, metadata, args, context=None):
        query = None if context is None else context.raw_query
        if isinstance(args[0], str):
            try:
                value = float(args[0])
            except:
                raise ArgumentParserException(
                    f"""Error parsing float argument '{metadata["name"]}' from string {repr(args[0])}""",
                    query=query,
                )
        elif isinstance(args[0], StringActionParameter):
            try:
                value = float(args[0].string)
            except:
                raise ArgumentParserException(
                    f"""Error parsing float argument '{metadata["name"]}' from {repr(args[0].string)}""",
                    position=args[0].position,
                    query=query,
                )
        elif isinstance(args[0], ExpandedActionParameter):
            try:
                value = float(args[0].value)
            except:
                raise ArgumentParserException(
                    f"""Error parsing float argument '{metadata["name"]}' from link {repr(args[0].link.encode())}""",
                    position=args[0].position,
                    query=query,
                )
        elif isinstance(args[0], ActionParameter):
            raise ArgumentParserException(
                f"Unsupported action parameter type {repr(args[0])} in FloatArgumentParser",
                position=args[0].position,
                query=query,
            )
        else:
            try:
                value = float(args[0])
            except:
                raise ArgumentParserException(
                    f"""Error parsing float argument '{metadata["name"]}' from object {repr(args[0])}""",
                    query=query,
                )

        return value, (value, metadata), args[1:]


class BooleanArgumentParser(ArgumentParser):
    def parse(self, metadata, args):
        return (
            dict(
                y=True,
                yes=True,
                n=False,
                no=False,
                t=True,
                true=True,
                f=False,
                false=False,
            ).get(str(args[0]).lower(), False),
            args[1:],
        )

    def parse_meta(self, metadata, args, context=None):
        query = None if context is None else context.raw_query

        def to_bool(x):
            return dict(
                y=True,
                yes=True,
                n=False,
                no=False,
                t=True,
                true=True,
                f=False,
                false=False,
            ).get(str(x).lower(), False)

        if isinstance(args[0], str):
            try:
                value = to_bool(args[0])
            except:
                raise ArgumentParserException(
                    f"""Error parsing boolean argument '{metadata["name"]}' from string {repr(args[0])}""",
                    query=query,
                )
        elif isinstance(args[0], StringActionParameter):
            try:
                value = to_bool(args[0].string)
            except:
                raise ArgumentParserException(
                    f"""Error parsing boolean argument '{metadata["name"]}' from value {repr(args[0].string)}""",
                    position=args[0].position,
                    query=query,
                )
        elif isinstance(args[0], ExpandedActionParameter):
            try:
                value = to_bool(args[0].value)
            except:
                raise ArgumentParserException(
                    f"""Error parsing boolean argument '{metadata["name"]}' from link {repr(args[0].link.encode())}""",
                    position=args[0].position,
                    query=query,
                )
        elif isinstance(args[0], ActionParameter):
            raise ArgumentParserException(
                f"Unsupported action parameter type {repr(args[0])} in BooleanArgumentParser",
                position=args[0].position,
                query=query,
            )
        else:
            try:
                value = dict(
                    y=True,
                    yes=True,
                    n=False,
                    no=False,
                    t=True,
                    true=True,
                    f=False,
                    false=False,
                ).get(str(args[0]).lower(), False)
            except:
                raise ArgumentParserException(
                    f"""Error parsing boolean argument '{metadata["name"]}' from object {repr(args[0])}""",
                    query=query,
                )

        return value, (value, metadata), args[1:]


class ListArgumentParser(ArgumentParser):
    def parse(self, metadata, args):
        return args, []

    def parse_meta(self, metadata, args, context=None):
        query = None if context is None else context.raw_query

        value = []
        for x in args:
            if isinstance(x, str):
                value.append(x)
            elif isinstance(x, StringActionParameter):
                value.append(x.string)
            elif isinstance(x, ExpandedActionParameter):
                value.append(x.value)
            elif isinstance(x, ActionParameter):
                raise ArgumentParserException(
                    f"Unsupported action parameter type {repr(x)} in ListArgumentParser",
                    position=x.position,
                    query=query,
                )
            else:
                raise ArgumentParserException(
                    f"Unsupported action parameter object {repr(x)} in ListArgumentParser",
                    query=query,
                )

        return value, (value, metadata), []


class ArgvArgumentParser(ListArgumentParser):
    is_argv = True


GENERIC_AP = ArgumentParser()
CONTEXT_AP = ContextArgumentParser()
INT_AP = IntArgumentParser()
FLOAT_AP = FloatArgumentParser()
BOOLEAN_AP = BooleanArgumentParser()
LIST_AP = ListArgumentParser()
ARGV_AP = ArgvArgumentParser()


def identifier_to_label(identifier):
    """Tries to convert an identifier to a more human readable text label.
    Replaces underscores by spaces and may do other tweaks.
    """
    txt = identifier.replace("_", " ")
    txt = txt.replace(" id", "ID")
    txt = dict(url="URL").get(txt, txt)
    txt = txt[0].upper() + txt[1:]
    return txt

def callable_hash(f):
    import hashlib
    h = hashlib.md5()
    h.update(f.__code__.co_code)
    separator=b"__sep3024325ab2a7__" #random string of bytes
    for x in f.__code__.co_consts:
        try:
            h.update(pickle.dumps(x))
        except:
            continue
        h.update(separator)

    sig = inspect.signature(f)
    for argname in list(sig.parameters):
        h.update(argname.encode('utf-8'))
        h.update(separator)
        p = sig.parameters[argname]
        if p.default != inspect.Parameter.empty:
            try:
                h.update(pickle.dumps(x))
            except:
                continue
        h.update(separator)

    return h.hexdigest()

def command_metadata_from_callable(f, has_state_argument=True, attributes=None):
    """Extract command metadata structure from a callable.
    Function interprets function name, document string, argument names and annotations into command metadata.
    """
    name = f.__name__
    doc = f.__doc__
    module = f.__module__
    annotations = f.__annotations__
    if attributes is None:
        attributes = dict(ns=None)
    if doc is None:
        doc = ""
    arguments = []
    sig = inspect.signature(f)
    for argname in list(sig.parameters):
        arg = dict(name=argname, label=identifier_to_label(argname))
        if argname == "context":
            p = sig.parameters[argname]
            if p.default != inspect.Parameter.empty:
                if p.default is not None:
                    raise Exception(
                        f"Default value for context in command {name} should be None"
                    )
                arg["default"] = p.default
            if argname in annotations:
                raise Exception(
                    f"Context in command {name} should not have type annotations"
                )
            arg["multiple"] = False
            arg["optional"] = True
            arg["type"] = "context"
            arg["editor"] = "ignore"
            arguments.append(arg)
            continue

        arg_type = None
        if argname in annotations:
            arg_annotation = annotations[argname]
            if type(arg_annotation) == type:
                arg_type = arg_annotation.__name__

        p = sig.parameters[argname]
        if p.default != inspect.Parameter.empty:
            arg["default"] = p.default
            arg["optional"] = True
            if arg_type is None and p.default is not None:
                arg_type = type(p.default).__name__
        else:
            arg["optional"] = False
        arg["multiple"] = p.kind is inspect.Parameter.VAR_POSITIONAL
        arg_editor = None
        if arg_type is not None:
            arg_editor = arg_type
        arg["type"] = arg_type
        arg["editor"] = arg_editor
        arguments.append(arg)

    if has_state_argument and len(arguments):
        state_argument = arguments[0]
        state_argument["pass_state"] = state_argument["name"] == "state"
        arguments = arguments[1:]
    else:
        state_argument = None

    return CommandMetadata(
        name=name,
        label=identifier_to_label(name),
        module=module,
        doc=doc,
        state_argument=state_argument,
        arguments=arguments,
        attributes=attributes,
        version=callable_hash(f)
    )


def argument_parser_from_command_metadata(command_metadata):
    """Create argument parser from command metadata"""
    ap = SequenceArgumentParser()
    for arg in command_metadata.arguments:
        if arg.get("multiple", False):
            ap += ARGV_AP
            break
        arg_type = arg.get("type")
        ap += dict(
            context=CONTEXT_AP,
            str=GENERIC_AP,
            int=INT_AP,
            float=FLOAT_AP,
            bool=BOOLEAN_AP,
            list=LIST_AP,
        ).get(arg_type, GENERIC_AP)
    return ap


class CommandExecutable(object):
    """Wrapper around a registered command
    Adapts arbitrary function to be used as a command.
    Function needs to be described by a command metadata structure and accompanied by an argument parser.
    This decodes all the arguments, executes the command and (if needed) wraps the result as a State.
    """

    def __init__(self, f, metadata, argument_parser):
        self.f = f
        self.metadata = metadata
        self.argument_parser = argument_parser

    def inner_id(self):
        return id(self.f)

    def parse_argv(self, args, context=None):
        query = None if context is None else context.raw_query
        args = list(args)
        try:
            position = args[-1].position
        except:
            position = None
        for i, a in list(enumerate(self.metadata.arguments))[len(args) :]:
            try:
                position = args[i].position
            except:
                position = None
            if not a.get("multiple", False) and a["name"] != "context":
                if "default" in a:
                    args.append(a["default"])
                else:
                    raise ArgumentParserException(
                        f"Expected '{a['name']}' argument for '{self.metadata.name}', no default",
                        position=position,
                        query=query,
                    )
        try:
            argv, argmeta, remainder = self.argument_parser.parse_meta(
                self.metadata.arguments, args, context=context
            )
        except ArgumentParserException as e:
            raise ArgumentParserException(
                f"{e.original_message} while executing '{self.metadata.name}'",
                position=e.position,
                query=query,
            ) from e
        if len(remainder) != 0:
            position = (
                remainder[0].position
                if isinstance(remainder[0], ActionParameter)
                else None
            )
            raise ArgumentParserException(
                f"Too many arguments for '{self.metadata.name}': {repr(remainder)}",
                position=position,
                query=query,
            )
        return argv, argmeta

    def __call__(self, state, *args, context=None):
        argv, argmeta = self.parse_argv(args, context=context)
        state_arg = state if self.metadata.state_argument["pass_state"] else state.get()
        result = self.f(state_arg, *argv)
        if isinstance(result, State):
            result.arguments = argmeta
            return result
        else:
            state.arguments = argmeta
            return state.with_data(result)


class FirstCommandExecutable(CommandExecutable):
    """Wrapper around a registered first command"""

    def __call__(self, state, *args, context=None):
        argv, argmeta = self.parse_argv(args, context=context)
        result = self.f(*argv)
        if isinstance(result, State):
            result.arguments = argmeta
            return result
        else:
            state.arguments = argmeta
            return state.with_data(result)


def command(*arg, **kwarg):
    """Register a callable as a command.
    Callable is expected to take a state data as a first argument.

    This function typically can be used as a decorator.
    As a decorator it can be used directly (@command) or it can have parameters,
    e.g. @command(ns="MyNameSpace")
    """
    if len(arg) == 1:
        assert callable(arg[0])
        f = arg[0]
        if "ns" not in kwarg:
            kwarg["ns"] = "root"
        metadata = command_metadata_from_callable(f, attributes=kwarg)
        command_registry().register_command(f, metadata)
        return f
    else:
        assert len(arg) == 0
        return lambda f, attributes=kwarg: command(f, **attributes)


def first_command(*arg, **kwarg):
    """Register a callable as a command.
    Unlike in command(), callable is expected NOT to take a state data as a first argument.
    Thus first_command can be a first command in the query - does not require a state to be applied on.
    However, first_command is a perfectly valid command, so it can as well be used inside the query and
    then the state passed to the command is ignored (and not passed to f).
    This typically can be used as a decorator.
    """
    if len(arg) == 1:
        assert callable(arg[0])
        f = arg[0]
        if "ns" not in kwarg:
            kwarg["ns"] = "root"
        metadata = command_metadata_from_callable(
            f, has_state_argument=False, attributes=kwarg
        )
        command_registry().register_command(f, metadata)
        return f
    else:
        assert len(arg) == 0
        return lambda f, attributes=kwarg: first_command(f, **attributes)



from io import StringIO, BytesIO
from tempfile import TemporaryDirectory
from pathlib import Path
import os

from keras.models import (
    Model,
    model_from_json,
    model_from_yaml,
    load_model,
    clone_model,
)
from keras.utils import plot_model

from liquer.state_types import StateType, register_state_type
from liquer.constants import mimetype_from_extension

from liquer.commands import command, first_command
from liquer.parser import encode, decode
from liquer.query import evaluate
from liquer.state import State
from tensorflow.python.keras.engine.functional import Functional

class KerasModelStateType(StateType):
    def identifier(self):
        return "kerasmodel"

    def default_extension(self):
        return "h5"

    def is_type_of(self, data):
        return isinstance(data, Model)

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)
        if extension == "json":
            output = StringIO()
            output.write(data.to_json())
            return output.getvalue().encode("utf-8"), mimetype
        elif extension == "yaml":
            output = StringIO()
            output.write(data.to_yaml())
            return output.getvalue().encode("utf-8"), mimetype
        elif extension in ("h5", "hdf5"):
            with TemporaryDirectory() as tmpdir:
                path = Path(tmpdir) / f"data.{extension}"
                data.save(str(path))
                b=path.read_bytes()
                return b, mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by kerasmodel type."
            )

    def from_bytes(self, b: bytes, extension=None):
        if extension is None:
            extension = self.default_extension()

        if extension == "json":
            return model_from_json(b)
        elif extension == "yaml":
            return model_from_yaml(b)
        elif extension in ["h5", "hdf5"]:
            with TemporaryDirectory() as tmpdir:
                path = Path(tmpdir) / f"data.{extension}"
                with open(path, "wb") as f:
                    f.write(b)
                model = load_model(f.name)
            return model
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by kerasmodel type."
        )

    def data_characteristics(self, data):
        s = StringIO()
        summary=keras_summary(data)
        return dict(description=f"Keras model", summary=summary)

    def copy(self, data):
        print ("KERAS MODEL COPY")
        model = clone_model(data)
        model.set_weights(data.get_weights())
        return model


KERASMODEL_STATE_TYPE = KerasModelStateType()
register_state_type(Model, KERASMODEL_STATE_TYPE)
register_state_type(Functional, KERASMODEL_STATE_TYPE)


@command
def keras_plot_model(
    model,
    show_shapes: bool = False,
    show_layer_names: bool = True,
    rankdir: str = "TB",
    expand_nested: bool = False,
    dpi: int = 96,
):
    "Keras plot model as png"
    print("************* keras_plot_model")
    assert isinstance(model, Model)
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir) / "keras_model.png"
        plot_model(
            model,
            to_file=str(path),
            show_shapes=show_shapes,
            show_layer_names=show_layer_names,
            rankdir=rankdir,
            expand_nested=expand_nested,
            dpi=96,
        )
        b = open(path, "rb").read()
        print("keras_plot_model OK")
        return b
    print("keras_plot_model failed")


@command
def keras_summary(model):
    "Keras model summary"
    s = StringIO()
    model.summary(print_fn=lambda x, s=s: s.write(str(x) + "\n"))
    summary=s.getvalue()
    return summary



import logging
from flask import Blueprint, jsonify, redirect, send_file, request, make_response, abort
from liquer.query import evaluate
from liquer.state_types import encode_state_data, state_types_registry
from liquer.commands import command_registry
from liquer.state import get_vars
from liquer.cache import get_cache
from liquer.store import get_store, KeyNotFoundStoreException
import io
import traceback

app = Blueprint("liquer", __name__, static_folder="static")
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@app.route("/", methods=["GET", "POST"])
@app.route("/index.html", methods=["GET", "POST"])
def index():
    """Link to a LiQuer main service page"""
    return redirect("/liquer/static/index.html")


@app.route("/info.html", methods=["GET", "POST"])
def info():
    """Info page"""
    return """
<html>
    <head>
        <title>LiQuer</title>
    </head>
    <body>
        <h1>LiQuer server</h1>
        For more info, see the <a href="https://github.com/orest-d/liquer">repository</a>.
    </body>    
</html>
"""


def response(state):
    """Create flask response from a State"""
    b, mimetype, type_identifier = encode_state_data(state.get(), extension=state.extension)
    filename = state.metadata.get("filename")
    if filename is None:
        filename = state_types_registry().get(type_identifier).default_filename()
    r = make_response(b)

    r.headers.set("Content-Type", mimetype)
    if mimetype not in [
        "application/json",
        "text/plain",
        "text/html",
        "text/csv",
        "image/png",
        "image/svg+xml",
    ]:
        r.headers.set("Content-Disposition", "attachment", filename=filename)
    return r


@app.route("/q/<path:query>")
def serve(query):
    """Main service for evaluating queries"""
    return response(evaluate(query))


@app.route("/submit/<path:query>")
def detached_serve(query):
    """Main service for evaluating queries"""
    from liquer.pool import evaluate_in_background

    evaluate_in_background(query)
    return jsonify(dict(status="OK", message="Submitted", query=query))


@app.route("/api/cache/get/<path:query>")
def cache_get(query):
    """Get cached metadata"""
    state = get_cache().get(query)
    if state is None:
        abort(404)
    return response(state)


@app.route("/api/cache/meta/<path:query>")
def cache_get_metadata(query):
    """Get cached metadata"""
    metadata = get_cache().get_metadata(query)
    if metadata == False:
        metadata = dict(query=query, status="not available", cached=False)
    return jsonify(metadata)


@app.route("/api/cache/meta/<path:query>", methods=["POST"])
def cache_store_metadata(query):
    """Store metadata in cache.
    Allows to use liquer server as a remote cache.
    """
    metadata = request.get_json(force=True)
    try:
        result_code = get_cache().store_metadata(metadata)
        result = dict(
            query=query, result=result_code, status="OK", message="OK", traceback=""
        )
    except Exception as e:
        result = dict(
            query=query,
            result=False,
            status="ERROR",
            message=str(e),
            traceback=traceback.format_exc(),
        )

    return jsonify(result)


@app.route("/api/cache/remove/<path:query>")
def cache_remove(query):
    """interface to cache remove"""
    r = get_cache().remove(query)
    return jsonify(dict(query=query, removed=r))


@app.route("/api/cache/contains/<path:query>")
def cache_contains(query):
    """interface to cache contains"""
    contains = get_cache().contains(query)
    return jsonify(dict(query=query, cached=contains))


@app.route("/api/cache/keys.json")
def cache_keys():
    """interface to cache keys"""
    keys = dict(keys=list(get_cache().keys()))
    return jsonify(keys)


@app.route("/api/cache/clean")
def cache_clean():
    """interface to cache clean"""
    get_cache().clean()
    n = len(list(get_cache().keys()))
    keys = dict(status="OK", message=f"Cache cleaned, {n} keys left")
    return jsonify(keys)


@app.route("/api/commands.json")
def commands():
    """Returns a list of commands in json format"""
    return jsonify(command_registry().as_dict())


@app.route("/api/debug-json/<path:query>")
def debug_json(query):
    """Debug query - returns metadata from a state after a query is evaluated"""
    state = evaluate(query)
    state_json = state.as_dict()
    return jsonify(state_json)


@app.route("/api/build", methods=["POST"])
def build():
    """Build a query from a posted decoded query (list of lists of strings).
    Result is a dictionary with encoded query and link.
    """
    from liquer.parser import encode

    query = encode(request.get_json(force=True)["ql"])
    link = (
        get_vars().get("server", "http://localhost")
        + get_vars().get("api_path", "/q/")
        + query
    )
    return jsonify(dict(query=query, link=link, message="OK", status="OK"))


@app.route("/api/register_command/<data>", methods=["GET"])
def register_command(data):
    """Remote command registration service.
    This has to be enabled by liquer.commands.enable_remote_registration()

    WARNING: Remote command registration allows to deploy arbitrary python code on LiQuer server,
    therefore it is a HUGE SECURITY RISK and it only should be used if other security measures are taken
    (e.g. on localhost or intranet where only trusted users have access).
    This is on by default on Jupyter server extension.
    """
    return jsonify(command_registry().register_remote_serialized(data.encode("ascii")))


@app.route("/api/register_command/", methods=["POST"])
def register_command1():
    """Remote command registration service.
    This has to be enabled by liquer.commands.enable_remote_registration()

    WARNING: Remote command registration allows to deploy arbitrary python code on LiQuer server,
    therefore it is a HUGE SECURITY RISK and it only should be used if other security measures are taken
    (e.g. on localhost or intranet where only trusted users have access).
    This is on by default on Jupyter server extension.
    """
    data = request.get_data()
    return jsonify(command_registry().register_remote_serialized(data))


@app.route("/api/store/data/<path:query>", methods=["GET"])
def store_get(query):
    """Get data from store. Equivalent to Store.get_bytes.
    Content type (MIME) is obtained from the metadata.
    """
    store = get_store()
    try:
        metadata = store.get_metadata(query)
        mimetype = metadata.get("mimetype", "application/octet-stream")
        r = make_response(store.get_bytes(query))
        r.headers.set("Content-Type", mimetype)
        return r
    except:
        response = jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )
        response.status="404"
        return response

@app.route("/web/<path:query>", methods=["GET"])
def web_store_get(query):
    """Shortcut to the 'web' directory in the store.
    Similar to /store/data/web, except the index.html is automatically added if query is a directory.
    The 'web' directory hosts web applications and visualization tools, e.g. liquer-pcv or liquer-gui. 
    """
    store = get_store()
    try:
        query="web/"+query
        if query.endswith("/"):
            query+="index.html"
        if store.is_dir(query):
            query+="/index.html"
        metadata = store.get_metadata(query)
        mimetype = metadata.get("mimetype", "application/octet-stream")
        r = make_response(store.get_bytes(query))
        r.headers.set("Content-Type", mimetype)
        return r
    except:
        return jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )

@app.route("/api/store/data/<path:query>", methods=["POST"])
def store_set(query):
    """Set data from store. Equivalent to Store.store.
    Unlike store method, which stores both data and metadata in one call,
    the api/store/data POST only stores the data. The metadata needs to be set in a separate POST of api/store/metadata
    either before or after the api/store/data POST.
    """
    store = get_store()
    try:
        metadata = store.get_metadata(query)
    except KeyNotFoundStoreException:
        metadata={}
    try:        
        data = request.get_data()
        store.store(query, data, metadata)
        return jsonify(dict(query=query, message="Data stored", status="OK"))
    except:
        response = jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )
        response.status="404"
        return response


@app.route("/api/store/metadata/<path:query>", methods=["GET"])
def store_get_metadata(query):
    store = get_store()
    metadata = store.get_metadata(query)
    return jsonify(metadata)


@app.route("/api/store/metadata/<path:query>", methods=["POST"])
def store_set_metadata(query):
    store = get_store()
    try:
        metadata = request.get_json(force=True)
        store.store_metadata(query, metadata)
        return jsonify(dict(query=query, message="Metadata stored", status="OK"))
    except:
        response = jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )
        response.status="404"
        return response

@app.route("/api/stored_metadata/<path:query>", methods=["GET"])
def get_stored_metadata(query):
    """Get metadata stored in a store or cache"""
    import liquer.tools
    metadata = liquer.tools.get_stored_metadata(query)
    return jsonify(metadata)

@app.route("/api/store/remove/<path:query>")
def store_remove(query):
    store = get_store()
    try:
        store.remove(query)
        return jsonify(dict(query=query, message=f"Removed {query}", status="OK"))
    except:
        return jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )


@app.route("/api/store/removedir/<path:query>")
def store_removedir(query):
    store = get_store()
    try:
        store.removedir(query)
        return jsonify(
            dict(query=query, message=f"Removed directory {query}", status="OK")
        )
    except:
        return jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )


@app.route("/api/store/contains/<path:query>")
def store_contains(query):
    store = get_store()
    try:
        contains = store.contains(query)
        return jsonify(
            dict(
                query=query, message=f"Contains {query}", contains=contains, status="OK"
            )
        )
    except:
        return jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )


@app.route("/api/store/is_dir/<path:query>")
def store_is_dir(query):
    store = get_store()
    try:
        is_dir = store.is_dir(query)
        return jsonify(
            dict(
                query=query, message=f"Is directory {query}", is_dir=is_dir, status="OK"
            )
        )
    except:
        return jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )


@app.route("/api/store/keys")
def store_keys():
    store = get_store()
    try:
        keys = store.keys()
        return jsonify(
            dict(query=None, message=f"Keys obtained", keys=keys, status="OK")
        )
    except:
        return jsonify(dict(query=None, message=traceback.format_exc(), status="ERROR"))


@app.route("/api/store/listdir/<path:query>")
def store_listdir(query):
    store = get_store()
    try:
        listdir = store.listdir(query)
        return jsonify(
            dict(query=query, message=f"Keys obtained", listdir=listdir, status="OK")
        )
    except:
        return jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )


@app.route("/api/store/makedir/<path:query>")
def store_makedir(query):
    store = get_store()
    try:
        store.makedir(query)
        return jsonify(dict(query=query, message=f"Makedir succeeded", status="OK"))
    except:
        return jsonify(
            dict(query=query, message=traceback.format_exc(), status="ERROR")
        )



from liquer.store import get_store, Store, KeyNotFoundStoreException, StoreException, join_key, parent_key
from liquer.context import get_context
from liquer.parser import parse
from liquer.constants import Status
from liquer.metadata import Metadata

from copy import deepcopy
import traceback
from yaml import load, dump
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

class RecipeException(Exception):
    def __init__(self, message, key=None):
        self.original_message = message
        if key is not None:
            message += f":\n  key: '{key}'"

        super().__init__(message)
        self.key = key

class RecipeRegistry(object):
    """Registry of recipes"""
    def __init__(self):
        self.recipe_dictionary = {}
    def register(self, recipe):
        if recipe.recipe_type() in self.recipe_dictionary:
            print(f"WARNING: Recipe type '{recipe.recipe_type()}' already registered")
        self.recipe_dictionary[recipe.recipe_type()] = recipe
    def from_dict(self, d):
        if type(d) != dict:
            raise Exception("Dictionary expected as recipe definition for registered recipes")
        if "type" not in d:
            raise Exception("Recipe definition is lacking a type")
        if d["type"] not in self.recipe_dictionary:
            t=d["type"]
            raise Exception(f"Recipe type '{t}' not registered")
        return self.recipe_dictionary[d["type"]].from_dict(d)

_recipe_registry = None


def recipe_registry():
    """Returns the global recipe registry (singleton)"""
    global _recipe_registry
    if _recipe_registry is None:
        _recipe_registry = RecipeRegistry()
    return _recipe_registry

def register_recipe(recipe):
    """Function to register new recipe type.
    """
    recipe_registry().register(recipe)

class Recipe:
    def __init__(self,d):
        if type(d) != dict:
            raise Exception("Dictionary expected as recipe definition")
        if "type" in d:
            if d["type"]!=self.recipe_type():
                t=d["type"]
                raise Exception(f"Recipe {self.recipe_type()} received definition with type='{t}'")
        self.data=d
    
    @classmethod
    def recipe_type(self):
        return "empty"

    @classmethod
    def from_dict(cls,d):
        return cls(d)
    
    def recipe_name(self):
        return self.data.get("recipe_name","")

    def provides(self):
        return self.data.get("provides",[])

    def can_create(self,name):
        return name in self.provides()

    def metadata(self, key):
        raise RecipeException("Recipe undefined (metadata)",key=key)
      
    def make(self, key, context=None):
        raise RecipeException("Recipe undefined (make)",key=key)

    def is_volatile(self):
        return self.data.get("volatile", False)

def resolve_recipe_definition(r, directory, metadata):
    if type(r) == str:
        try:
            query = parse(r)
            filename = query.filename()
            return dict(
                type="query",
                query=r,
                CWD = directory,
                filename=filename,
                provides=[filename]
            )
        except:
            metadata.warning(f"Can't resolve recipe '{r}'", traceback=traceback.format_exc())
            traceback.print_exc()
    elif isinstance(r, dict):
        if r.get("type") in (None, "query") and "query" in r:
            try:
                query = parse(r["query"])
                filename = r.get("filename", query.filename())
                title = r.get("title", filename)
                description = r.get("description", f'Generated from query: {r["query"]}')
                rkey =  join_key(directory, filename)
                return dict(
                    type="query",
                    query=r["query"],
                    title=title,
                    description=description,
                    CWD = directory,
                    filename=filename,
                    provides=[filename]
                )
            except:
                metadata.warning(f"Can't resolve query recipe", traceback=traceback.format_exc())
                traceback.print_exc()
    else:
        print(f"Unsupported recipe type: {type(r)}")
    if "filename" in r and "provides" not in r:
        r["provides"] = [r["filename"]]
    return r

class QueryRecipe(Recipe):
    @classmethod
    def recipe_type(self):
        return "query"

    @classmethod
    def from_dict(cls,d):
        return cls(d)
        
    def metadata(self, key):
        metadata = Metadata(dict())
        if "title" in self.data:
            metadata.metadata["title"] = self.data["title"]
        if "description" in self.data:
            metadata.metadata["description"] = self.data["description"]
        metadata.query = self.data["query"]
        metadata.key=key
        return metadata.as_dict()
      
    def make(self, key, store=None, context=None):
        context = get_context(context)
        if store is None:
            store = context.store()
        context.evaluate(
            self.data["query"],
            store_key=key,
            store_to=store,
        )

register_recipe(QueryRecipe)

class RecipeStore(Store):
    def __init__(self, store, recipes=None, context=None):
        self.substore = store
        self.substore.parent_store=self
        self._recipes = {} if recipes is None else recipes
        self.context = context

    def sync(self):
        self.substore.sync()

    def get_context(self):
        if self.context is None:
            return get_context()
        else:
            return self.context.new_empty()

    def with_context(self, context):
        return RecipeStore(self.substore, recipes=self.recipes, context=context)

    def mount_recipe(self, key, recipe):
        self._recipes[key] = recipe
        return self

    def ignore(self, key):
        return False

    def make(self, key):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't make it")
        query = self.recipes().get(key)
        if query is None:
            raise KeyNotFoundStoreException(
                f"Key {key} not found, recipe unknown", key=key, store=self
            )
        target_resource_directory = self.parent_key(key)
        target_file = self.key_name(key)
        self.get_context().evaluate(
            query,
            store_key=key,
            store_to=self,
        )
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def recipes(self):
        return self._recipes

    def recipe_metadata(self, key):
        return {}

    def is_supported(self, key):
        if self.ignore(key):
            return False
        return self.substore.is_supported(key)

    def get_bytes(self, key):
        if self.ignore(key):
            return None
        if self.substore.contains(key):
            return self.substore.get_bytes(key)
        self.make(key)
        return self.substore.get_bytes(key)

    def get_metadata(self, key):
        if self.ignore(key):
            raise KeyNotFoundStoreException(key=key, store=self)
        if self.substore.contains(key):
            return self.substore.get_metadata(key)
        if self.is_dir(key):
            return self.finalize_metadata(
                self.default_metadata(key=key, is_dir=True), key=key, is_dir=True
            )
        if key in self.recipes():
            metadata = self.recipe_metadata(key)
            try:
                sub_metadata = self.substore.get_metadata(key)
                if sub_metadata is not None:
                    metadata.update(sub_metadata)
            except:
                pass
            return self.finalize_metadata(metadata, key=key, is_dir=False)
        raise KeyNotFoundStoreException(key=key, store=self)

    def store(self, key, data, metadata):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't store into it")
        self.substore.store(
            key, data, self.finalize_metadata(metadata, key=key, is_dir=True, data=data)
        )
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't store metadata into it")
        self.substore.store_metadata(key, metadata)
        self.on_metadata_changed(key)

    def remove(self, key):
        self.substore.remove(key)
        self.on_removed(key)

    def removedir(self, key):
        self.substore.removedir(key)
        self.on_removed(key)

    def contains(self, key):
        if self.ignore(key):
            return False
        if self.substore.contains(key):
            return True
        for k in self.recipes():
            if k == key or k.startswith(key + "/"):
                return True
        return False

    def is_dir(self, key):
        if self.ignore(key):
            return False
        if self.substore.is_dir(key):
            return True
        for k in self.recipes():
            if k == key:
                return False
            if k.startswith(key + "/"):
                return True
        return False

    def keys(self):
        return [
            key
            for key in sorted(set(self.substore.keys()).union(self.recipes().keys()))
            if not self.ignore(key)
        ]

    def listdir(self, key):
        if self.ignore(key):
            return []
        d = set(self.substore.listdir(key) or [])
        key_split = key.split("/")
        if len(key_split) == 1 and key_split[0] == "":
            key_split = []
        key_depth = len(key_split)
        for k in self.recipes().keys():
            if k.startswith(key + "/") or key in (None, ""):
                v = k.split("/")
                d.add(v[key_depth])
        return [key for key in sorted(d) if not self.ignore(key)]

    def makedir(self, key):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't makedir")
        self.substore.makedir(key)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="r", buffering=-1):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't openbin")
        return self.substore.openbin(key, mode=mode, buffering=buffering)

    def __str__(self):
        return f"Recipe store on ({self.substore})"

    def __repr__(self):
        return f"RecipeStore({repr(self.substore)}, recipes={repr(self.recipes())})"


class OldRecipeSpecStore(RecipeStore):
    RECIPES_FILE = "recipes.yaml"
    LOCAL_RECIPES = "RECIPES"
    STATUS_FILE = "recipes_status.txt"

    def __init__(self, store, context=None):
        RecipeStore.__init__(self, store, recipes=None, context=context)
        self.recipes_info = {}
        self.update_recipes()
        self.update_all_status_files()

    def sync(self):
        self.update_recipes()
        self.update_all_status_files()
        self.substore.sync()

    def update_all_status_files(self):
        if self.STATUS_FILE is not None:
            for dir_key in set(self.parent_key(key) for key in self.recipes().keys()):
                self.create_status(dir_key)

    def ignore(self, key):
        if key is None:
            return True
        return any(x.startswith(".") for x in key.split("/"))

    def recipe_metadata(self, key):
        metadata = deepcopy(self.recipes_info.get(key, {}))
        metadata["status"] = Status.RECIPE.value
        if metadata.get("query") is not None:
            metadata["has_recipe"] = True
            if self.recipes_info[key].get("recipes_directory") == self.LOCAL_RECIPES:
                metadata["recipes_directory"] = ""
            else:
                metadata["recipes_directory"] = self.recipes_info[key].get(
                    "recipes_directory"
                )
        return metadata

    def make(self, key):
        print(f"### MAKE {key}")
        super().make(key)
        metadata = self.substore.get_metadata(key)
        status = metadata.get("status", Status.RECIPE.value)
        fileinfo = metadata["fileinfo"]
        metadata.update(self.recipe_metadata(key))
        metadata["status"] = status
        metadata["fileinfo"] = fileinfo
        self.substore.store_metadata(key, metadata)
        self.on_metadata_changed(key)

    def update_recipes(self):
        import yaml

        recipes = {}
        for key in self.substore.keys():
            spec = None
            if self.key_name(key) == self.RECIPES_FILE and not self.substore.is_dir(
                key
            ):
                spec = yaml.load(self.substore.get_bytes(key), Loader=Loader)
            recipes_key = key
            if spec is not None:
                parent = self.parent_key(key)
                for directory, items in spec.items():
                    for r in items:
                        if type(r) == str:
                            try:
                                query = parse(r)
                                filename = query.filename()
                                parent = self.parent_key(key)
                                if len(parent) > 0 and not parent.endswith("/"):
                                    parent += "/"
                                rkey = (
                                    f"{parent}{filename}"
                                    if directory == self.LOCAL_RECIPES
                                    else f"{parent}{directory}/{filename}"
                                )
                                recipes[rkey] = r
                                self.recipes_info[rkey] = dict(
                                    query=r,
                                    title=filename,
                                    description="",
                                    recipes_key=recipes_key,
                                    recipes_directory=directory,
                                )
                            except:
                                traceback.print_exc()
                        elif isinstance(r, dict):
                            try:
                                query = parse(r["query"])
                                filename = r.get("filename", query.filename())
                                title = r.get("title", filename)
                                description = r.get("description", r["query"])
                                parent = self.parent_key(key)
                                if len(parent) > 0 and not parent.endswith("/"):
                                    parent += "/"
                                rkey = (
                                    f"{parent}{filename}"
                                    if directory == self.LOCAL_RECIPES
                                    else f"{parent}{directory}/{filename}"
                                )
                                recipes[rkey] = r["query"]
                                self.recipes_info[rkey] = dict(
                                    query=r["query"],
                                    title=title,
                                    description=description,
                                    recipes_key=recipes_key,
                                    recipes_directory=directory,
                                )
                            except:
                                traceback.print_exc()
                        else:
                            print(f"Unsupported recipe type: {type(r)}")
        self._recipes = recipes
        return recipes

    def create_status_text(self, dir_key):
        txt = ""
        if self.substore.is_dir(dir_key):
            for d in self.listdir(dir_key):
                key = f"{dir_key}/{d}" if len(dir_key) else d
                if d == self.STATUS_FILE:
                    continue
                if not self.is_dir(key):
                    metadata = self.get_metadata(key)
                    if metadata is None:
                        txt += "%-14s %-30s %s\n" % ("MISSING", d, "Missing metadata")
                    else:
                        try:
                            t = metadata.get("created")
                            if t in ("", None):
                                t = metadata["updated"]
                            time = util.format_datetime(util.to_datetime(t))
                        except:
                            time = ""
                        status = metadata.get("status", Status.NONE.value)
                        message = metadata.get("message", "").strip()
                        if status == Status.READY.value:
                            try:
                                message = metadata["data_characteristics"][
                                    "description"
                                ]
                            except:
                                pass
                        if "\n" in message:
                            txt += "%-20s %-14s %-32s|" % (time, status, d)
                            txt += "\n=============================================================\n"
                            txt += message
                            txt += "\n=============================================================\n\n"
                        else:
                            txt += "%-20s %-14s %-32s| %s\n" % (
                                time,
                                status,
                                d,
                                message,
                            )
                        trace = ""
                        for entry in metadata.get("log", []) + metadata.get(
                            "child_log", []
                        ):
                            tb = entry.get("traceback")
                            if tb is not None:
                                if len(tb):
                                    if "timestamp" in entry:
                                        trace += f"Time:    {entry['timestamp']}"
                                    if "origin" in entry:
                                        trace += f"Origin:  {entry['origin']}"
                                    if "message" in entry:
                                        trace += f"Message: {entry['message']}"
                                    trace += "\n"
                                    trace += tb
                                    tb += "\n------------------------\n"
                        if len(trace):
                            txt += "\n### TRACEBACK ###############################################\n"
                            txt += trace
                            txt += "\n#############################################################\n\n"
        return txt

    def create_status(self, key):
        if self.key_name(key) != self.STATUS_FILE:
            if not self.is_dir(key):
                key = self.parent_key(key)
            status_key = f"{key}/{self.STATUS_FILE}" if len(key) else self.STATUS_FILE
            data = self.create_status_text(key).encode("utf-8")
            self.substore.store(
                status_key,
                data,
                dict(
                    type_identifier="text",
                    status=Status.SIDE_EFFECT.value,
                    title=f"Status of '{key}'",
                    description="This file is generated automatically by the recipe store",
                ),
            )

    def on_metadata_changed(self, key):
        super().on_metadata_changed(key)
        if self.STATUS_FILE is not None:
            if self.key_name(key) == self.RECIPES_FILE:
                self.update_all_status_files()
            elif self.key_name(key) != self.STATUS_FILE:
                self.create_status(key)

    def on_data_changed(self, key):
        super().on_data_changed(key)
        if self.key_name(key) == self.RECIPES_FILE:
            self.update_recipes()
            self.update_all_status_files()
        else:
            self.create_status(key)

    def on_removed(self, key):
        super().on_removed(key)
        if self.key_name(key) == self.RECIPES_FILE:
            self.update_recipes()
            self.update_all_status_files()
        else:
            self.create_status(key)

class NewRecipeSpecStore(Store):
    RECIPES_FILE = "recipes.yaml"
    LOCAL_RECIPES = "RECIPES"
    STATUS_FILE = "recipes_status.txt"

    def __init__(self, store):
        self.substore = store
        self.substore.parent_store=self
        self._recipes = {}
        self.update_recipes()
        self.update_all_status_files()

    def sync(self):
        self.update_recipes()
        self.update_all_status_files()
        self.substore.sync()

    def recipes(self):
        return self._recipes

    def make(self, key):
        print(f"### MAKE {key}")
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't make it")

        recipe = self.recipes().get(key)
        if recipe is None:
            raise KeyNotFoundStoreException(
                f"Key {key} not found, recipe unknown", key=key, store=self
            )
        recipe.make(key, store=self.substore)

        metadata = self.substore.get_metadata(key)
        status = metadata.get("status", Status.RECIPE.value)
        fileinfo = metadata["fileinfo"]
        metadata.update(self.recipe_metadata(key))
        metadata["status"] = status
        metadata["fileinfo"] = fileinfo
        self.substore.store_metadata(key, metadata)
        self.on_data_changed(key)
        self.on_metadata_changed(key)


    def recipe_metadata(self, key):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't make it")

        recipe = self.recipes().get(key)
        if recipe is None:
            raise KeyNotFoundStoreException(
                f"Key {key} not found, recipe unknown", key=key, store=self
            )
        metadata = recipe.metadata(key)
        metadata["status"] = Status.RECIPE.value
        metadata["has_recipe"] = True

        return metadata

    def is_supported(self, key):
        if self.ignore(key):
            return False
        return self.substore.is_supported(key)

    def get_bytes(self, key):
        if self.ignore(key):
            return None
        if self.substore.contains(key):
            return self.substore.get_bytes(key)
        self.make(key)
        return self.substore.get_bytes(key)

    def get_metadata(self, key):
        if self.ignore(key):
            raise KeyNotFoundStoreException(key=key, store=self)
        if self.substore.contains(key):
            return self.substore.get_metadata(key)
        if self.is_dir(key):
            return self.finalize_metadata(
                self.default_metadata(key=key, is_dir=True), key=key, is_dir=True
            )
        if key in self.recipes():
            metadata = self.recipe_metadata(key)
            try:
                sub_metadata = self.substore.get_metadata(key)
                if sub_metadata is not None:
                    metadata.update(sub_metadata)
            except:
                pass
            return self.finalize_metadata(metadata, key=key, is_dir=False)
        raise KeyNotFoundStoreException(key=key, store=self)

    def store(self, key, data, metadata):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't store into it")
        self.substore.store(
            key, data, self.finalize_metadata(metadata, key=key, is_dir=True, data=data)
        )
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def store_metadata(self, key, metadata):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't store metadata into it")
        self.substore.store_metadata(key, metadata)
        self.on_metadata_changed(key)

    def remove(self, key):
        self.substore.remove(key)
        self.on_removed(key)

    def removedir(self, key):
        self.substore.removedir(key)
        self.on_removed(key)

    def contains(self, key):
        if self.ignore(key):
            return False
        if self.substore.contains(key):
            return True
        for k in self.recipes():
            if k == key or k.startswith(key + "/"):
                return True
        return False

    def is_dir(self, key):
        if self.ignore(key):
            return False
        if self.substore.is_dir(key):
            return True
        for k in self.recipes():
            if k == key:
                return False
            if k.startswith(key + "/"):
                return True
        return False

    def keys(self):
        return [
            key
            for key in sorted(set(self.substore.keys()).union(self.recipes().keys()))
            if not self.ignore(key)
        ]

    def listdir(self, key):
        if self.ignore(key):
            return []
        d = set(self.substore.listdir(key) or [])
        key_split = key.split("/")
        if len(key_split) == 1 and key_split[0] == "":
            key_split = []
        key_depth = len(key_split)
        for k in self.recipes().keys():
            if k.startswith(key + "/") or key in (None, ""):
                v = k.split("/")
                d.add(v[key_depth])
        return [key for key in sorted(d) if not self.ignore(key)]

    def makedir(self, key):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't makedir")
        self.substore.makedir(key)
        self.on_data_changed(key)
        self.on_metadata_changed(key)

    def openbin(self, key, mode="r", buffering=-1):
        if self.ignore(key):
            raise Exception(f"Key {key} is ignored, can't openbin")
        return self.substore.openbin(key, mode=mode, buffering=buffering)

    def __str__(self):
        return f"Recipe spec store n ({self.substore})"

    def __repr__(self):
        return f"RecipeSpecStore({repr(self.substore)})"


    def update_all_status_files(self):
        if self.STATUS_FILE is not None:
            for dir_key in set(parent_key(key) for key in self.recipes().keys()):
                self.create_status(dir_key)

    def ignore(self, key):
        if key is None:
            return True
        return any(x.startswith(".") for x in key.split("/"))


    def update_recipes(self):
        import yaml

        recipes = {}
        for key in self.substore.keys():
            spec = None
            if self.key_name(key) == self.RECIPES_FILE and not self.substore.is_dir(
                key
            ):
                spec = yaml.load(self.substore.get_bytes(key), Loader=Loader)
            recipes_key = key
            metadata = Metadata(self.substore.get_metadata(key))
            if spec is not None:
                parent = parent_key(key)
                for directory, items in spec.items():
                    for i, r in enumerate(items):
                        cwd = parent if directory == self.LOCAL_RECIPES else join_key(parent, directory)
                        d = resolve_recipe_definition(r, cwd, metadata)
                        d["recipe_name"] = self.to_root_key(recipes_key)+f":{directory}[{i}]"+d.get("filename","")
                        recipe = recipe_registry().from_dict(d)
                        for name in recipe.provides():
                            key = join_key(cwd, name)
                            recipes[key]=recipe
        self._recipes = recipes
        return recipes

    def create_status_text(self, dir_key):
        txt = ""
        if self.substore.is_dir(dir_key):
            for d in self.listdir(dir_key):
                key = f"{dir_key}/{d}" if len(dir_key) else d
                if d == self.STATUS_FILE:
                    continue
                if not self.is_dir(key):
                    metadata = self.get_metadata(key)
                    if metadata is None:
                        txt += "%-14s %-30s %s\n" % ("MISSING", d, "Missing metadata")
                    else:
                        try:
                            t = metadata.get("created")
                            if t in ("", None):
                                t = metadata["updated"]
                            time = util.format_datetime(util.to_datetime(t))
                        except:
                            time = ""
                        status = metadata.get("status", Status.NONE.value)
                        message = metadata.get("message", "").strip()
                        if status == Status.READY.value:
                            try:
                                message = metadata["data_characteristics"][
                                    "description"
                                ]
                            except:
                                pass
                        if "\n" in message:
                            txt += "%-20s %-14s %-32s|" % (time, status, d)
                            txt += "\n=============================================================\n"
                            txt += message
                            txt += "\n=============================================================\n\n"
                        else:
                            txt += "%-20s %-14s %-32s| %s\n" % (
                                time,
                                status,
                                d,
                                message,
                            )
                        trace = ""
                        for entry in metadata.get("log", []) + metadata.get(
                            "child_log", []
                        ):
                            tb = entry.get("traceback")
                            if tb is not None:
                                if len(tb):
                                    if "timestamp" in entry:
                                        trace += f"Time:    {entry['timestamp']}"
                                    if "origin" in entry:
                                        trace += f"Origin:  {entry['origin']}"
                                    if "message" in entry:
                                        trace += f"Message: {entry['message']}"
                                    trace += "\n"
                                    trace += tb
                                    tb += "\n------------------------\n"
                        if len(trace):
                            txt += "\n### TRACEBACK ###############################################\n"
                            txt += trace
                            txt += "\n#############################################################\n\n"
        return txt

    def create_status(self, key):
        if self.key_name(key) != self.STATUS_FILE:
            if not self.is_dir(key):
                key = self.parent_key(key)
            status_key = f"{key}/{self.STATUS_FILE}" if len(key) else self.STATUS_FILE
            data = self.create_status_text(key).encode("utf-8")
            self.substore.store(
                status_key,
                data,
                dict(
                    type_identifier="text",
                    status=Status.SIDE_EFFECT.value,
                    title=f"Status of '{key}'",
                    description="This file is generated automatically by the recipe store",
                ),
            )

    def on_metadata_changed(self, key):
        super().on_metadata_changed(key)
        if self.STATUS_FILE is not None:
            if self.key_name(key) == self.RECIPES_FILE:
                self.update_all_status_files()
            elif self.key_name(key) != self.STATUS_FILE:
                self.create_status(key)

    def on_data_changed(self, key):
        super().on_data_changed(key)
        if self.key_name(key) == self.RECIPES_FILE:
            self.update_recipes()
            self.update_all_status_files()
        else:
            self.create_status(key)

    def on_removed(self, key):
        super().on_removed(key)
        if self.key_name(key) == self.RECIPES_FILE:
            self.update_recipes()
            self.update_all_status_files()
        else:
            self.create_status(key)

class RecipeSpecStore(NewRecipeSpecStore):
    pass


import json
from liquer.state_types import (
    type_identifier_of,
    copy_state_data,
    data_characteristics
)
from liquer.constants import mimetype_from_extension

from copy import deepcopy
from liquer.parser import QueryException, Position

_vars = None


class EvaluationException(QueryException):
    pass


def get_vars():
    global _vars
    if _vars is None:
        _vars = {}
    return _vars


def vars_clone():
    """Return a copy of state variables dictionary
    This is used to set the initial content of State.vars for a newly created state.
    """
    return deepcopy(get_vars())


def set_var(name, value):
    """Set initial value of a state variable
    This is used to configure initial content of the state variables.

    It should be done before any query is evaluated.
    Calls at a later stage may lead to unintended behaviour like
    cache inconsistencies (query result may depend on initial value of a state variable,
    but this is not reflected in a query string).
    """
    get_vars()
    _vars[name] = value


class State(object):
    def __init__(self, data=None, metadata=None, context=None):
        self.data = data
        self.exception = None
        self.context = context
        self.metadata = (
            metadata
            if metadata is not None
            else dict(
                query="",
                sources=[],
                log=[],
                is_error=False,
                vars=vars_clone(),
                filename=None,
                extension=None,
                message="",
                commands=[],
                extended_commands=[],
                type_identifier=type_identifier_of(data),
                caching=True,
                attributes={},
                data_characteristics=data_characteristics(data)
            )
        )

    def next_state(self):
        state = self.__class__()
        state = state.from_dict(self.as_dict()).with_data(None)
        return state

    @property
    def query(self):
        return self.metadata["query"]

    @query.setter
    def query(self, value):
        self.metadata["query"] = value

    @property
    def is_error(self):
        return self.metadata["is_error"]

    @is_error.setter
    def is_error(self, value):
        self.metadata["is_error"] = value

    @property
    def vars(self):
        return self.metadata["vars"]

    @vars.setter
    def vars(self, value):
        self.metadata["vars"] = value

    @property
    def type_identifier(self):
        return self.metadata["type_identifier"]

    @type_identifier.setter
    def type_identifier(self, value):
        self.metadata["type_identifier"] = value

    @property
    def extension(self):
        return self.metadata.get("extension")

    @extension.setter
    def extension(self, value):
        self.metadata["extension"] = value

    #    def with_caching(self, caching=True):
    #        """Enables or disables caching for this state"""
    #        # TODO: Make sure caching is propagated to dependent states
    #        # TODO: Examples and documentation
    #        print(f"QUERY {self.query}\nWITH CACHING {caching}")
    #        if self.context is not None:
    #            self.context.caching = caching
    #            print(f"set context {id(self.context)} caching {self.context.caching}")
    #        self.metadata["caching"] = caching
    #        return self

    def with_data(self, data):
        """Set the data"""
        self.data = data
        self.metadata["type_identifier"] = type_identifier_of(data)
        self.metadata["data_characteristics"] = data_characteristics(data)
        return self

    def with_source(self, source):
        """Add sources to the state"""
        self.metadata["sources"] = [source] + self.metadata["sources"]
        return self

    def is_volatile(self):
        return self.metadata.get("attributes", {}).get("volatile", False)

    def set_volatile(self, flag):
        self.metadata["attributes"]["volatile"] = flag
        return self

    def get(self):
        """Get data from the state"""
        if self.is_error:
            tb = "\n".join(m.get("traceback", "") for m in self.metadata["log"])
            print(tb)
            position = None
            query = None
            for entry in self.metadata["log"]:
                if entry.get("kind") == "error":
                    position = Position.from_dict(entry.get("position"))
                    query = entry.get("query")

            if self.exception is None:
                raise EvaluationException(
                    tb + "\n" + self.metadata["message"], position=position, query=query
                )
            else:
                raise self.exception

        return self.data

    def log_command(self, qv, number):
        """Log a command"""
        if self.context is None:
            self.metadata["log"].append(
                dict(kind="command", qv=qv, command_number=number)
            )
        else:
            self.context.log_action(qv, number)
        return self

    def log_error(self, message):
        """Log an error message"""
        if self.context is None:
            self.metadata["log"].append(dict(kind="error", message=message))
            self.is_error = True
            self.metadata["message"] = message
        else:
            self.context.error(message)
        return self

    def log_warning(self, message):
        """Log a warning message"""
        if self.context is None:
            self.metadata["log"].append(dict(kind="warning", message=message))
            self.metadata["message"] = message
        else:
            self.context.warning(message)
        return self

    def log_exception(self, message, traceback):
        """Log an exception"""
        if self.context is None:
            self.metadata["log"].append(
                dict(kind="error", message=message, traceback=traceback)
            )
            self.is_error = True
            self.metadata["message"] = message
        else:
            self.context.exception(message, traceback)
        return self

    def log_info(self, message):
        """Log a message (info)"""
        if self.context is None:
            self.metadata["log"].append(dict(kind="info", message=message))
            self.metadata["message"] = message
        else:
            self.context.info(message)
        return self

    def as_dict(self):
        """Represent state metadata as a dictionary - suitable for json serialization
        State data is NOT part of the returned dictionary.
        """
        return deepcopy(self.metadata)

    def mimetype(self):
        """Return mime type of the data"""
        
        return self.metadata.get("mimetype", mimetype_from_extension(self.extension))

    def from_dict(self, metadata):
        """Fill state by valueas from a dictionary"""
        if isinstance(metadata, self.__class__):
            metadata = metadata.metadata
        self.metadata = deepcopy(metadata)
        return self

    def has_flag(self, name):
        return self.metadata["vars"].get(name) == True

    def clone(self):
        """Clone the state including the deep copy of the data"""
        state = self.__class__()
        state = state.from_dict(self.as_dict())
        state.data = copy_state_data(self.data)
        return state

    def with_filename(self, filename):
        """set filename"""
        self.metadata["filename"] = filename
        print("State with filename",filename)
        if "." in filename:
            self.extension = filename.split(".")[-1].lower()
            self.metadata["mimetype"] =  mimetype_from_extension(self.extension)
        print("  extension",self.extension)
        print("  mimetype ",self.metadata["mimetype"])
        return self
